/*****************************************************************************/
/*
                               qdLogStats.c

Quick & Dirty LOG STATisticS.

Utility to extract very elementary statistics from Web server common/combined
format log files.  A number of filters allow subsets of the log contents to be
selected.  These filters are simple sort-of regular expressions, not
case-sensitive (deliberately), can contain wildcards (such as asterisks (*),
question marks (?), and percent signs (%)) as well as sort-of regular
expressions (such as the range [a-z]).  THERE IS NO WAY (that I know of) TO
ESCAPE THESE RESERVED CHARACTERS!  This functionality uses decc$match_wild()
function.  When using filters it is necessary to provide wildcard 'match-alls'
at the start and/or end of the string if the searched-for string does not fall
at the start and/or end of the field.  Hence to find all records having a field
containing "fred" anywhere in the string the following must be used

  "*fred*"

Special constructs allow more complex expressions to be built up.  Combinations
of required and excluded strings may be specified in the one expression.  When
a string begins with a "+{" it *must* be present for the record NOT to be
filtered out.  If it begins "-{" then it must *not* be present.  Such
specifications must be terminated with a matching closure "}".  For example,
with the following expression

  "+{*test*}-{*fred*}"

the record field must contain a string "test" somewhere but cannot contain the
string "fred" anywhere.  The expression

  "+{*one*}+{*two*}"

specifies that the strings "one" and "two" must be present in the field but not
necessarily in any particular order.  An expression

  "-{*three*}-{*four*}"

excludes the field if it contains either of the two fields anywhere in it.

Matching log file records can be viewed as processed using the
/VIEW=[type] qualifier, where <type> is ALL, MATCH (default) or NOMATCH.

For fields that are commonly URL-encoded (or for any field that can usefully be
URL-encoded to disguise the actual path) the /DECODE qualifier specifies that
the field should be URL-decoded before filter matching.  If the /DECODE
qualifier is used without keywords all three of these fields are decoded,
otherwise any combination of PATH, QUERY and REFERER may be used to selectively
decode that field.  For example /DECODE=(PATH,QUERY).  Remember that decoding
uses CPU cycles and adds processing time.

A progress indicator can be displayed using the /PROGRESS qualifier.  With this
a "+" is output for each file processed and a "." for each 1000 records in any
file.  An optional integer can be provides as /PROGRESS=integer to change this
record value.

Note that all selections are based on string matching.  Files and/or records
cannot be selected on the basis of being before or after a particular date for
instance.  Also that records are accepted or rejected in the order in which the
results are displayed, that is if a record is rejected against method it will
never be assessed against user-agent ... UNLESS then /ALL qualifier is applied. 
When /ALL is used comparison continues against all filter strings even if one
or more do not match.  This is useful for displaying the totals that match each
of multiple filter critera (not much use against a single one).

If /LOOKUP is enabled then dotted-decimal client host addresses have their
names resolved (if possible).  This name is displayed when log line is output. 
When enabled the resolved name is also used when matching against client hosts.

Although part of the WASD package there is no reason why this shouldn't be of
use with other packages if desired.


EXAMPLES
--------
Requires a foreign verb or such (shorter version might save keystrokes :^)

  $ QDLOG == "$CGI-BIN:[000000]QDLOGSTATS"
  $ QDLOGSTATS == "$CGI-BIN:[000000]QDLOGSTATS"

As the character sequences '/' then '*', and '*' then '/' are reserved in C,
the following examples have '@' substituted for '*'.  Please read '*' for all
instances of '@'!

1)  Records from September 1999.

    $ QDLOGSTATS HT_LOGS:@1999@.LOG /DATE="@/SEP/1999@"

2)  Records where the browser was an X-based Netscape Navigator

    $ QDLOGSTATS HT_LOGS:@.LOG /USERAGENT=@MOZILLA@X11@

3)  Records of POST method requests

    $ QDLOGSTATS HT_LOGS:@.LOG /METHOD=POST

4)  Records requesting a particular path

    $ QDLOGSTATS HT_LOGS:@.LOG /PATH="/cgi-bin/@"

5)  Records with a query string (second excludes those containing "grotty")

    $ QDLOGSTATS HT_LOGS:@.LOG /QUERY="%@"
    $ QDLOGSTATS HT_LOGS:@.LOG /QUERY="+{%@}-{@grotty@}"

6)  Records requesting a particular path, excluding the help script

    $ QDLOGSTATS HT_LOGS:@.LOG /PATH="+{/cgi-bin/@}-{/cgi-bin/help/@}"

7)  Select proxy records requesting (a) particular site(s)

    $ QDLOGSTATS HT_LOGS:@8080@.LOG /PATH="http://@.compaq.com@"
    $ QDLOGSTATS HT_LOGS:@8080@.LOG /METHOD=POST /PATH="http://@sex@.@/@"

8)  Records where the request was authenticated

    $ QDLOGSTATS HT_LOGS:@.LOG /AUTHUSER="-{-}"
    $ QDLOGSTATS HT_LOGS:@.LOG /AUTHUSER="DANIEL"

9)  Records where the request was authenticated, excluding DANIEL

    $ QDLOGSTATS HT_LOGS:@.LOG /AUTHUSER="-{-}-{DANIEL}"

10) Records with requests originating from (a) particular host(s)

    $ QDLOGSTATS HT_LOGS:@.LOG /CLIENT=131.185.250.1
    $ QDLOGSTATS HT_LOGS:@.LOG /LOOKUP /CLIENT=@.vsm.com.au


CGI INTERFACE
-------------
A CGI interface is available.  It basically parallels the CLI behaviour
described above. 

  http://the.host.name/cgi-bin/qdlogstats

So that it's not available for uncontrolled use the script *must* be subject to
authorization (i.e. have a non-empty REMOTE_USER).  For the WASD package this
may be enabled using the following HTTPD$AUTH rule.

  [whatever-realm]
  /cgi-bin/qdlogstats r+w,~siteadmin

For VMS Apache (CSWS) using SYSUAF authentication this might be configured
using the following entries in [CONF]HTTPD.CONF

  LoadModule auth_openvms_module
  /apache$common/modules/mod_auth_openvms.exe_alpha

  <Location /cgi-bin/qdlogstats>
  AuthType Basic
  AuthName "OpenVMS authentication"
  AuthUserOpenVMS On
  require valid-user
  </Location>

If there seems to be no relevant CGI variables the script displays a request
form that allows the entering of strings against the various fields of the log
file entries.  This form uses the GET method (putting form elements into the
query string) and so searches commonly or periodically made can be saved as
bookmarks if desired.

The account processing the script must have access to the log directory and
files.  This may be provided by the server account itself (i.e. the logs are
normally accessable) or the scripting account itself has access.  For instance,
with WASD on VMS V6.2 or later this could be provided using the /PERSONA
functionality and a HTTPD$MAP rule such as

  set /cgi-bin/qdlogstats script=as=SYSTEM

An alternative is to install the script executable with sufficient privileges
to access the files.  Normally this would only be SYSPRV but as the script
attempts to enable READALL as well so it *could* be installed with that if
necessary.  The following example DCL executed during server startup would
ensure such access.

  $ INSTALL ADD /PRIVILEGE=READALL CGI-BIN:[000000]QDLOGSTATS.EXE

For WASD 8.1 and later access control should not need to be changed.  For
non-WASD and WASD earlier than 8.1 the following should be applied.  If this is
provided it would also be prudent to control access to the executable from
arbitrary accounts at the CLI, and restricting it to the scripting account
using an ACL.  For example

  $ SET SECURITY CGI-BIN:[000000]QDLOGSTATS.EXE -
        /ACL=((IDENT=HTTP$SERVER,ACCESS=R+E),(IDENT=*,ACCESS=NONE)

The CGI version REQUIRES the following system-level logical name to locate the
log files for it.  If this name does not exist the CGI interface will not work. 
All log files must be available via this logical.  The logical name is defined
/SYSTEM to limit a possibly READALL privilege being used against arbitrary
paths.

  $ DEFINE /SYSTEM QDLOGSTATS_LOGS "HT_LOGS:"


QUALIFIERS
----------
/ALL                  compare all records to all filters
/AUTHUSER=            pattern (any authenticated username)
/BEFORE=              files (not records!) modified before VMS time
/CLIENT=              pattern (client host name or IP address)
/DATETIME=            pattern ("11/Jun/1999:14:08:49 +0930")
/DBUG                 turns on all "if (Debug)" statements
/DECODE[=keyword]     URL-decode PATH, QUERY and/or REFERER before filtering
/LOG                  show the name of the log file the records come from
/LOOKUP               lookup host name from dotted decimal address
/METHOD=              HTTP method ("GET", "POST", "HEAD", etc.)
/OUTPUT=              file specification
/PATH=                pattern (URL path component only)
/PROGRESS[=integer]   show progress during processing
/QUERY=               pattern (URL query component only)
/REFERER=             pattern (HTTP "Referer:" field, COMBINED only)
/REMOTEID=            pattern (anything in the RFC819 remote ident field)
/RESPONSE=            HTTP response status code pattern
/SINCE=               files (not records!) modified since VMS time
/SIZE=([MIN=|MAX=])   minimum and/or maximum response size  
/SOFTWAREID           (and /VERSION) display QDLOGSTATS and CGILIB versions
/USERAGENT=           pattern (HTTP "User-Agent:", COMBINED only)
/VIEW[=type]          display matching records, where <type> can be
                      ALL, MATCH (default) or NOMATCH


BUILD DETAILS
-------------
See BUILD_QDLOGSTATS.COM procedure.


COPYRIGHT
---------
Copyright (C) 2000-2003 Mark G.Daniel
This program, comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it
under the conditions of the GNU GENERAL PUBLIC LICENSE, version 2.


VERSION HISTORY (update SOFTWAREID as well!)
---------------
02-JAN-2003  MGD  v1.7.0, filtering on HTTP response status and response size
24-NOV-2002  MGD  v1.6.0, URL-decoding path/query/referer before filtering,
                          check for account SYSPRV before CLI activating
05-NOV-2002  MGD  v1.5.1, extend CGI log search to check for QDLOGSTATS_LOGS
                          then to fallback to HTTPD$LOG and HT_LOGS logicals
26-SEP-2002  MGD  v1.5.0, make CGI log search dependent on the QDLOGSTATS_LOGS
                          logical name being defined and log files within that
16-AUG-2002  MGD  v1.4.0, add log file last modification before/since,
                          some accomodations for CSWS 1.2,
                          bugfix; silly boy - what about HEAD method!
30-MAR-2002  MGD  v1.3.0, accumulate request methods
01-JUL-2001  MGD  v1.2.2, add 'SkipParameters' for direct OSU support
30-JAN-2001  MGD  v1.2.1, small improvement to lookup cache for noting
                          addresses that never resolve (and do it slowly)
10-JAN-2001  MGD  v1.2.0, lookup client host name
23-DEC-2000  MGD  v1.1.0, CGI interface (still pretty q&d)
14-NOV-2000  MGD  v1.0.0, initial development
*/
/*****************************************************************************/

#ifdef __ALPHA
#  define SOFTWAREID "QDLOGSTATS AXP-1.7.0"
#else
#  define SOFTWAREID "QDLOGSTATS VAX-1.7.0"
#endif

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <jpidef.h>
#include <prvdef.h>
#include <stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unixio.h>
#include <unixlib.h>

/* VMS-related header files */
#include <descrip.h>
#include <lnmdef.h>
#include <lib$routines.h>
#include <prvdef.h>
#include <rms.h>
#include <ssdef.h>
#include <starlet.h>
#ifndef __ALPHA
   /* "Compaq C V6.4-005 on OpenVMS VAX V7.3" seems to need this */
#  define SYS$GETMSG sys$getmsg
   int sys$getmsg (__unknown_params);
#endif
#include <stsdef.h>
#include <syidef.h>

/* TCP/IP-related header files */
#include <in.h>
#include <netdb.h>
#include <inet.h>

/* application related header files */
#include <cgilib.h>

#define BOOL int
#define true 1
#define false 0

#define VMSok(x) ((x) & STS$M_SUCCESS)
#define VMSnok(x) (!((x) & STS$M_SUCCESS))
 
#define FI_LI "QDLOGSTATS",__LINE__

/* used in modulus for every so many records progress */
#define PROGRESS_RECORDS 1000
/* wrap every so many characters */
#define PROGRESS_WRAP_CLI 80
#define PROGRESS_WRAP_CGI 60

/* maximum log line/record length */
#define MAX_LINE_LENGTH 2048

/* maximum names in host cache */
#define MAX_LOOKUP_CACHE 256

/* maximum times an address will attempt to be resolved */
#define MAX_LOOKUP_FAILS 2

#define DEFAULT_LOGS_LOGICAL "QDLOGSTATS_LOGS"
#define DEFAULT_LOGS2_LOGICAL "HTTPD$LOG"
#define DEFAULT_LOGS3_LOGICAL "HT_LOGS"
#define DEFAULT_LOGS_FILESPEC "*.LOG*;"
#define DEFAULT_LOGS_DIR_APACHE "APACHE$SPECIFIC:[LOGS]"
#define DEFAULT_LOGS_APACHE "*ACCESS_LOG*.;"
#define DEFAULT_LOGS_DIR_OSU  "WWW_ROOT:[SYSTEM]"
#define DEFAULT_LOGS_OSU "ACCESS*.LOG;"
#define DEFAULT_LOGS_DIR_WASD "HT_LOGS:"
#define DEFAULT_LOGS_WASD "*ACCESS*.LOG*;"

#define VIEW_NONE     0
#define VIEW_LOG      1
#define VIEW_ALL      2
#define VIEW_MATCH    3
#define VIEW_NOMATCH  4

char  CopyrightInfo [] =
"Copyright (C) 2000-2003 Mark G.Daniel.\n\
This software comes with ABSOLUTELY NO WARRANTY.\n\
This is free software, and you are welcome to redistribute it\n\
under the conditions of the GNU GENERAL PUBLIC LICENSE, version 2.";

char  Utility [] = "QDLOGSTATS";

BOOL  Debug,
      DoCgiStats,
      DoCliStats,
      FilterOnAll,
      LookupName,
      ShowLogFile,
      UrlDecodePath,
      UrlDecodeQuery,
      UrlDecodeReferer;

int  AccessProblemCount,
     AuthUserFilterAcceptCount,
     AuthUserFilterRejectCount,
     ClientFilterAcceptCount,
     ClientFilterRejectCount,
     CtimeBefore,
     CtimeSince,
     CombinedLogRecordCount,
     CommonLogRecordCount,
     DateTimeFilterRejectCount,
     DateTimeFilterAcceptCount,
     FileCountTotal,
     LastModifiedIgnoredCount,
     MethodFilterAcceptCount,
     MethodFilterRejectCount,
     MethodConnectCount,
     MethodDeleteCount,
     MethodGetCount,
     MethodHeadCount,
     MethodPostCount,
     MethodPutCount,
     MethodUnknownCount,
     PathFilterAcceptCount,
     PathFilterRejectCount,
     ProgressCount,
     ProgressWrap,
     QueryFilterAcceptCount,
     QueryFilterRejectCount,
     RecordCount,
     RecordCountTotal,
     SuspectRecordCountTotal,
     RequestCountTotal,
     RefererFilterAcceptCount,
     RefererFilterRejectCount,
     RemoteIdentFilterAcceptCount,
     RemoteIdentFilterRejectCount,
     ResponseMaxSize,
     ResponseMaxSizeAcceptCount,
     ResponseMaxSizeRejectCount,
     ResponseMinSize,
     ResponseMinSizeAcceptCount,
     ResponseMinSizeRejectCount,
     ShowProgress,
     StatusFilterAcceptCount,
     StatusFilterRejectCount,
     UserAgentFilterAcceptCount,
     UserAgentFilterRejectCount,
     ViewRecords;

int  StatusCodeCount [6];

float  ByteCountTotal;

char  *AuthUserFilterPtr,
      *CgiFormLogFileSpecPtr,
      *CgiLibEnvironmentPtr,
      *CgiScriptNamePtr,
      *CgiServerNamePtr,
      *ClientFilterPtr,
      *DateTimeFilterPtr,
      *LastModifiedBeforePtr,
      *LastModifiedSincePtr,
      *LookupNamePtr,
      *MethodFilterPtr,
      *OutputPtr,
      *PathFilterPtr,
      *QueryFilterPtr,
      *RefererFilterPtr,
      *RemoteIdentFilterPtr,
      *SearchDna,
      *ResponseFilterPtr,
      *UserAgentFilterPtr;

char  LogFileSpec [256],
      SoftwareID [64],
      StartDateTime [32];

$DESCRIPTOR (StartDateTimeFaoDsc, "!17%D\0");
$DESCRIPTOR (StartDateTimeDsc, StartDateTime);

/* required function prototypes */
BOOL FilterThisOut (char*, char*);
void CgiStats ();
void CgiForm ();
void CliStats ();
int DayMonthYear (char*);
char* DateTimeFilter (char*);
char* SysGetMsg (int);
int GetParameters ();
char* LookupHostName (char**);
void NeedsPrivilegedAccount ();
void OutputLogRecord (char*);
int ProcessLogFile (char*);
int ProcessLogRecord (char*);
int SearchFileSpec (char*);
int ShowHelp ();
int strsame (char*, char*, int);
char* SysTrnLnmLnmSystem (char*);

/*****************************************************************************/
/*
'argc' and 'argv' are only required to support CgiLibEnvironmentInit(), in
particular the OSU environment.
*/

main
(
int argc,
char *argv[]
)
{
   /*********/
   /* begin */
   /*********/

   sprintf (SoftwareID, "%s (%s)", SOFTWAREID, CGILIB_SOFTWAREID);

   /* always initialize the CGILIB so we can at least determine the mode */
   if (getenv ("QDLOGSTATS$DBUG")) Debug = true;
   CgiLibEnvironmentSetDebug (Debug);
   if (Debug) fprintf (stdout, "Content-Type: text/plain\n\n");

   GetParameters ();

   CgiLibEnvironmentInit (argc, argv, false);

   /* determine if we're at the command line or being used as a script */
   if (!CgiLibVarNull ("WWW_GATEWAY_INTERFACE"))
   {
      DoCliStats = true;
      CliStats ();
   }
   else
   {
      DoCgiStats = true;
      CgiStats ();
   }

   exit (SS$_NORMAL);
}

/*****************************************************************************/
/*
*/

void CliStats ()

{
   static char  ModifiedBetween [] =
      "Modified ..... between %11.11s and %11.11s\n";

   int  status,
        DurationHr,
        DurationMin,
        DurationSec;
   char  *cptr;
   char  AccessProblemString [256],
         AuthUserString [256],
         ByteString [256],
         ClientString [256],
         DateTimeString [256],
         IgnoredString [256],
         LastModifiedString [128],
         MethodString [256],
         PathString [256],
         QueryString [256],
         RefererString [256],
         RemoteIdentString [256],
         ResponseMaxSizeString [128],
         ResponseMinSizeString [128],
         ResponseString [128],
         UserAgentString [256];
   unsigned long  StartSeconds,
                  TimeSeconds;

   /*********/
   /* begin */
   /*********/

   NeedsPrivilegedAccount ();

   if (OutputPtr)
      if (!(stdout = freopen (OutputPtr, "w", stdout)))
         exit (vaxc$errno);

   if (!LogFileSpec[0])
   {
      ShowHelp ();
      exit (SS$_NORMAL);
   }

   if (LastModifiedSincePtr)
      if (!(CtimeSince = DayMonthYear (LastModifiedSincePtr)))
         exit (SS$_IVTIME);

   if (LastModifiedBeforePtr)
      if (!(CtimeBefore = DayMonthYear (LastModifiedBeforePtr)))
         exit (SS$_IVTIME);

   sys$fao (&StartDateTimeFaoDsc, NULL, &StartDateTimeDsc, 0);
   if (StartDateTime[0] == ' ') strcpy (StartDateTime, StartDateTime+1);
   fprintf (stdout, "%s, %s\n", SoftwareID, StartDateTime);
   time (&StartSeconds);

   if (LastModifiedSincePtr && LastModifiedBeforePtr)
      sprintf (LastModifiedString, ModifiedBetween,
               LastModifiedSincePtr, LastModifiedBeforePtr);
   else
   if (LastModifiedSincePtr)
      sprintf (LastModifiedString, ModifiedBetween,
               LastModifiedSincePtr, StartDateTime);
   else
   if (LastModifiedBeforePtr)
      sprintf (LastModifiedString, ModifiedBetween,
               StartDateTime, LastModifiedBeforePtr);
   else
      LastModifiedString[0] = '\0';

   ProgressWrap = PROGRESS_WRAP_CLI;

   /* make a guess at the log file defaults */
   if (getenv("HT_ROOT"))
   {
      SearchDna = DEFAULT_LOGS_DIR_WASD;
      cptr = DEFAULT_LOGS_WASD;
   }
   else
   if (getenv("WWW_ROOT"))
   {
      SearchDna = DEFAULT_LOGS_DIR_OSU ;
      cptr = DEFAULT_LOGS_OSU;
   }
   else
   if (getenv("APACHE$SPECIFIC"))
   {
      SearchDna = DEFAULT_LOGS_DIR_APACHE;
      cptr = DEFAULT_LOGS_APACHE;
   }
   else
   {
      SearchDna = DEFAULT_LOGS_FILESPEC;
      cptr = DEFAULT_LOGS_FILESPEC;
   }

   if (LogFileSpec[0]) cptr = LogFileSpec;

   status = SearchFileSpec (cptr);

   if (ShowProgress && FileCountTotal) fputs ("\n", stdout);

   if (VMSnok (status) && status != RMS$_NMF) exit (status);

   if (AccessProblemCount)
      sprintf (AccessProblemString, ", %d access failed", AccessProblemCount);
   else
      AccessProblemString[0] = '\0';

   if (LastModifiedIgnoredCount)
      sprintf (IgnoredString, ", %d ignored", LastModifiedIgnoredCount);
   else
      IgnoredString[0] = '\0';

   if (ByteCountTotal > 1000000000.0)
   {
      ByteCountTotal /= 1000000000.0;
      sprintf (ByteString, "GBytes ....... %.3f", ByteCountTotal);
   }
   else
   if (ByteCountTotal > 1000000.0)
   {
      ByteCountTotal /= 1000000.0;
      sprintf (ByteString, "MBytes ........ %.3f", ByteCountTotal);
   }
   else
   if (ByteCountTotal > 1000.0)
   {
      ByteCountTotal /= 1000.0;
      sprintf (ByteString, "KBytes ........ %.3f", ByteCountTotal);
   }
   else
      sprintf (ByteString, "Bytes ......... %.0f", ByteCountTotal);

   time (&TimeSeconds);
   TimeSeconds -= StartSeconds;
   DurationHr = TimeSeconds / 3600;
   TimeSeconds -= DurationHr * 3600;
   DurationMin = TimeSeconds / 60;
   DurationSec = TimeSeconds - DurationMin * 60;

   if (!ResponseMinSize)
      strcpy (ResponseMinSizeString, "<none>");
   else
   if (ResponseMinSize < 1000)
      sprintf (ResponseMinSizeString, "%d", ResponseMinSize); 
   else
   if (ResponseMinSize < 1000000)
      sprintf (ResponseMinSizeString, "%dk", ResponseMinSize / 1000); 
   else
      sprintf (ResponseMinSizeString, "%dM", ResponseMinSize / 1000000); 

   if (!ResponseMaxSize)
      strcpy (ResponseMaxSizeString, "<none>");
   else
   if (ResponseMaxSize < 1000)
      sprintf (ResponseMaxSizeString, "%d", ResponseMaxSize); 
   else
   if (ResponseMaxSize < 1000000)
      sprintf (ResponseMaxSizeString, "%dk", ResponseMaxSize / 1000); 
   else
      sprintf (ResponseMaxSizeString, "%dM", ResponseMaxSize / 1000000); 

   if (FilterOnAll)
   {
      if (!ClientFilterPtr)
         strcpy (ClientString, "<none>");
      else
         sprintf (ClientString, "%s  (%d nomatch, %d match)",
                  ClientFilterPtr, ClientFilterRejectCount,
                                   ClientFilterAcceptCount);
      if (!AuthUserFilterPtr)
         strcpy (AuthUserString, "<none>");
      else
         sprintf (AuthUserString, "%s  (%d nomatch, %d match)",
                  AuthUserFilterPtr, AuthUserFilterRejectCount,
                                     AuthUserFilterAcceptCount);
      if (!DateTimeFilterPtr)
         strcpy (DateTimeString, "<none>");
      else
         sprintf (DateTimeString, "%s  (%d nomatch, %d match)",
                  DateTimeFilterPtr, DateTimeFilterRejectCount,
                                     DateTimeFilterAcceptCount);
      if (!MethodFilterPtr)
         strcpy (MethodString, "<none>");
      else
         sprintf (MethodString, "%s  (%d nomatch, %d match)",
                  MethodFilterPtr, MethodFilterRejectCount,
                                   MethodFilterAcceptCount);
      if (!PathFilterPtr)
         strcpy (PathString, "<none>");
      else
         sprintf (PathString, "%s  (%d nomatch, %d match)",
                  PathFilterPtr, PathFilterRejectCount,
                                 PathFilterAcceptCount);
      if (!QueryFilterPtr)
         strcpy (QueryString, "<none>");
      else
         sprintf (QueryString, "  (%d nomatch, %d match)",
                  QueryFilterRejectCount, QueryFilterAcceptCount);
      if (!RefererFilterPtr)
         strcpy (RefererString, "<none>");
      else
         sprintf (RefererString, "%s  (%d nomatch, %d match)",
                  RefererFilterPtr, RefererFilterRejectCount,
                                    RefererFilterAcceptCount);
      if (!RemoteIdentFilterPtr)
         strcpy (RemoteIdentString, "<none>");
      else
         sprintf (RemoteIdentString, "%s  (%d nomatch, %d match)",
                  RemoteIdentFilterPtr, RemoteIdentFilterRejectCount,
                                        RemoteIdentFilterAcceptCount);
      if (!UserAgentFilterPtr)
         strcpy (UserAgentString, "<none>");
      else
         sprintf (UserAgentString, "%s  (%d nomatch, %d match)",
                  UserAgentFilterPtr, UserAgentFilterRejectCount,
                                      UserAgentFilterAcceptCount);
      if (!ResponseFilterPtr)
         strcpy (ResponseString, "<none>");
      else
         sprintf (ResponseString, "%s  (%d nomatch, %d match)",
                  ResponseFilterPtr, StatusFilterRejectCount,
                                     StatusFilterAcceptCount);
      if (ResponseMinSize)
      {
         for (cptr = ResponseMinSizeString; *cptr; cptr++);
         sprintf (cptr, "  (%d nomatch, %d match)",
                  ResponseMinSizeRejectCount, ResponseMinSizeAcceptCount);
      }
      if (ResponseMaxSize)
      {
         for (cptr = ResponseMaxSizeString; *cptr; cptr++);
         sprintf (cptr, "  (%d nomatch, %d match)",
                  ResponseMaxSizeRejectCount, ResponseMaxSizeAcceptCount);
      }
   }
   else
   {
      if (!ClientFilterPtr)
         strcpy (ClientString, "<none>");
      else
         sprintf (ClientString, "%s  (%d nomatch)",
                  ClientFilterPtr, ClientFilterRejectCount);
      if (!AuthUserFilterPtr)
         strcpy (AuthUserString, "<none>");
      else
         sprintf (AuthUserString, "%s  (%d nomatch)",
                  AuthUserFilterPtr, AuthUserFilterRejectCount);
      if (!DateTimeFilterPtr)
         strcpy (DateTimeString, "<none>");
      else
         sprintf (DateTimeString, "%s  (%d nomatch)",
                  DateTimeFilterPtr, DateTimeFilterRejectCount);
      if (!MethodFilterPtr)
         strcpy (MethodString, "<none>");
      else
         sprintf (MethodString, "%s  (%d nomatch)",
                  MethodFilterPtr, MethodFilterRejectCount);
      if (!PathFilterPtr)
         strcpy (PathString, "<none>");
      else
         sprintf (PathString, "%s  (%d nomatch)",
                  PathFilterPtr, PathFilterRejectCount);
      if (!QueryFilterPtr)
         strcpy (QueryString, "<none>");
      else
         sprintf (QueryString, "%s  (%d nomatch)",
                  QueryFilterPtr, QueryFilterRejectCount);
      if (!RefererFilterPtr)
         strcpy (RefererString, "<none>");
      else
         sprintf (RefererString, "%s  (%d nomatch)",
                  RefererFilterPtr, RefererFilterRejectCount);
      if (!RemoteIdentFilterPtr)
         strcpy (RemoteIdentString, "<none>");
      else
         sprintf (RemoteIdentString, "%s  (%d nomatch)",
                  RemoteIdentFilterPtr, RemoteIdentFilterRejectCount);
      if (!UserAgentFilterPtr)
         strcpy (UserAgentString, "<none>");
      else
         sprintf (UserAgentString, "%s  (%d nomatch)",
                  UserAgentFilterPtr, UserAgentFilterRejectCount);
      if (!ResponseFilterPtr)
         strcpy (ResponseString, "<none>");
      else
         sprintf (ResponseString, "%s  (%d nomatch)",
                  ResponseFilterPtr, StatusFilterRejectCount);
      if (ResponseMinSize)
      {
         for (cptr = ResponseMinSizeString; *cptr; cptr++);
         sprintf (cptr, "  (%d nomatch)",
                  ResponseMinSizeRejectCount);
      }
      if (ResponseMaxSize)
      {
         for (cptr = ResponseMaxSizeString; *cptr; cptr++);
         sprintf (cptr, "  (%d nomatch)",
                  ResponseMaxSizeRejectCount);
      }
   }

   fprintf (stdout,
"\n\
Log ........... %s  (%d matched%s%s)\n\
%s\
Client ........ %s\n\
Remote ID ..... %s\n\
Auth User ..... %s\n\
Date/Time ..... %s\n\
Method ........ %s\n\
Path .......... %s\n\
Query ......... %s\n\
Referer ....... %s\n\
User Agent .... %s\n\
\n\
Response ...... %s\n\
Size Min ...... %s\n\
Size Max ...... %s\n\
\n\
Duration ...... %02d:%02d:%02d\n\
Records ....... %d  (%d suspect, %d common, %d combined)\n\
Requests ...... %d\n\
Methods ....... CONNECT:%d  DELETE:%d  GET:%d  HEAD:%d  POST:%d  PUT:%d  ?:%d\n\
Responses ..... 1nn:%d  2nn:%d  3nn:%d  4nn:%d  5nn:%d  ?:%d\n\
%s\n",
      LogFileSpec, FileCountTotal, IgnoredString, AccessProblemString,
      LastModifiedString,
      ClientString,
      RemoteIdentString,
      AuthUserString,
      DateTimeString,
      MethodString,
      PathString,
      QueryString,
      RefererString,
      UserAgentString,
      ResponseString,
      ResponseMinSizeString,
      ResponseMaxSizeString,
      DurationHr, DurationMin, DurationSec,
      RecordCountTotal, SuspectRecordCountTotal,
      CommonLogRecordCount, CombinedLogRecordCount,
      RequestCountTotal,
      MethodConnectCount,
      MethodDeleteCount,
      MethodGetCount,
      MethodHeadCount,
      MethodPostCount,
      MethodPutCount,
      MethodUnknownCount,
      StatusCodeCount[1],
      StatusCodeCount[2],
      StatusCodeCount[3],
      StatusCodeCount[4],
      StatusCodeCount[5],
      StatusCodeCount[0],
      ByteString);
}      

/*****************************************************************************/
/*
*/

void CgiStats ()

{
#ifndef PRV$M_READALL
#define PRV$M_READALL 0x0008
#endif

   static unsigned long  PrvMask [2] =
      { (unsigned long)PRV$M_SYSPRV | (unsigned long)PRV$M_READALL, 0 };

   static char  ModifiedBetween [] =
"<TR><TH ALIGN=right>Modified:&nbsp;</TH>\
<TD>between %11.11s and %11.11s</TD></TR>\n";

   BOOL  ShowProgressOnPage;
   int  status,
        DurationHr,
        DurationMin,
        DurationSec;
   char  *cptr,
         *ByteStringPtr;
   char  AccessProblemString [256],
         AuthUserString [256],
         ByteString [16],
         ClientString [256],
         DateTimeString [256],
         IgnoredString [256],
         LastModifiedString [128],
         MethodString [256],
         PathString [256],
         QueryString [256],
         RefererString [256],
         RemoteIdentString [256],
         ResponseMaxSizeString [128],
         ResponseMinSizeString [128],
         ResponseString [128],
         UserAgentString [256],
         VmsMessage [512];
   unsigned long  StartSeconds,
                  TimeSeconds;

   /*********/
   /* begin */
   /*********/

   sprintf (SoftwareID, "%s (%s)", SOFTWAREID, CgiLibEnvironmentVersion());

   /* ensure one of the required log directory logicals is defined */
   if (!(SearchDna = SysTrnLnmLnmSystem(DEFAULT_LOGS_LOGICAL)))
      if (!(SearchDna = SysTrnLnmLnmSystem(DEFAULT_LOGS2_LOGICAL)))
         if (!(SearchDna = SysTrnLnmLnmSystem(DEFAULT_LOGS3_LOGICAL)))
         {
            CgiLibResponseError (FI_LI, 0, "Logical name "
                                           DEFAULT_LOGS_LOGICAL
                                           " is not defined.");
            return;
         }
   
   sys$fao (&StartDateTimeFaoDsc, NULL, &StartDateTimeDsc, 0);
   if (StartDateTime[0] == ' ') strcpy (StartDateTime, StartDateTime+1);
   time (&StartSeconds);

   cptr = CgiLibVar ("WWW_REMOTE_USER");
   if (!cptr[0])
   {
      CgiLibResponseError (FI_LI, 0, "Authorization mandatory.");
      return;
   }

   CgiLibEnvironmentRecordOut ();
   CgiLibEnvironmentPtr = CgiLibEnvironmentName();
   CgiServerNamePtr = CgiLibVar ("WWW_SERVER_NAME");

   CgiFormLogFileSpecPtr = CgiLibVarNull ("WWW_FORM_LOGFILESPEC");
   if (!CgiFormLogFileSpecPtr)
   {
      CgiForm ();
      return;
   }

   LastModifiedSincePtr = CgiLibVarNull ("WWW_FORM_LOGMODSINCE");
   if (LastModifiedSincePtr && *(long*)LastModifiedSincePtr == 'dd-m')
      LastModifiedSincePtr = NULL;
   if (LastModifiedSincePtr)
      if (!(CtimeSince = DayMonthYear (LastModifiedSincePtr)))
      {
         CgiLibResponseError (FI_LI, SS$_IVTIME, "modified since.");
         exit (SS$_NORMAL);
      }

   LastModifiedBeforePtr = CgiLibVarNull ("WWW_FORM_LOGMODBEFORE");
   if (LastModifiedBeforePtr && *(long*)LastModifiedBeforePtr == 'dd-m')
      LastModifiedBeforePtr = NULL;
   if (LastModifiedBeforePtr)
      if (!(CtimeBefore = DayMonthYear (LastModifiedBeforePtr)))
      {
         CgiLibResponseError (FI_LI, SS$_IVTIME, "modified before.");
         exit (SS$_NORMAL);
      }

   if (LastModifiedSincePtr && LastModifiedBeforePtr)
      sprintf (LastModifiedString, ModifiedBetween,
               LastModifiedSincePtr, LastModifiedBeforePtr);
   else
   if (LastModifiedSincePtr)
      sprintf (LastModifiedString, ModifiedBetween,
               LastModifiedSincePtr, StartDateTime);
   else
   if (LastModifiedBeforePtr)
      sprintf (LastModifiedString, ModifiedBetween,
               StartDateTime, LastModifiedBeforePtr);
   else
      LastModifiedString[0] = '\0';

   /* get the filter components from the form */
   ClientFilterPtr = CgiLibVar ("WWW_FORM_CLIENT");
   if (!ClientFilterPtr[0])
      ClientFilterPtr = NULL;
   else
      /*
         Strictly speaking I shouldn't fool about with CGILIB variables
         without copying the contents but I'm just forcing it to lower
         case as FilterThisOut() will be expecting it this way!
      */
      for (cptr = ClientFilterPtr; *cptr; cptr++) *cptr = tolower(*cptr);
   RemoteIdentFilterPtr = CgiLibVar ("WWW_FORM_REMOTEID");
   if (!RemoteIdentFilterPtr[0])
      RemoteIdentFilterPtr = NULL;
   else
      for (cptr = RemoteIdentFilterPtr; *cptr; cptr++) *cptr = tolower(*cptr);
   AuthUserFilterPtr = CgiLibVar ("WWW_FORM_AUTHUSER");
   if (!AuthUserFilterPtr[0])
      AuthUserFilterPtr = NULL;
   else
      for (cptr = AuthUserFilterPtr; *cptr; cptr++) *cptr = tolower(*cptr);
   DateTimeFilterPtr = CgiLibVarNull ("WWW_FORM_DATETIME");
   if (DateTimeFilterPtr)
      DateTimeFilterPtr = DateTimeFilter (DateTimeFilterPtr);
   if (DateTimeFilterPtr)
      for (cptr = DateTimeFilterPtr; *cptr; cptr++) *cptr = tolower(*cptr);
   MethodFilterPtr = CgiLibVar ("WWW_FORM_METHOD");
   if (!MethodFilterPtr[0])
      MethodFilterPtr = CgiLibVar ("WWW_FORM_METHODLIST");
   if (!MethodFilterPtr[0])
      MethodFilterPtr = NULL;
   else
      for (cptr = MethodFilterPtr; *cptr; cptr++) *cptr = tolower(*cptr);
   PathFilterPtr = CgiLibVar ("WWW_FORM_PATH");
   if (!PathFilterPtr[0])
      PathFilterPtr = NULL;
   else
      for (cptr = PathFilterPtr; *cptr; cptr++) *cptr = tolower(*cptr);
   cptr = CgiLibVar ("WWW_FORM_DECODEPATH");
   if (cptr[0]) UrlDecodePath = true; else UrlDecodePath = false;
   QueryFilterPtr = CgiLibVar ("WWW_FORM_QUERY");
   if (!QueryFilterPtr[0])
      QueryFilterPtr = NULL;
   else
      for (cptr = QueryFilterPtr; *cptr; cptr++) *cptr = tolower(*cptr);
   cptr = CgiLibVar ("WWW_FORM_DECODEQUERY");
   if (cptr[0]) UrlDecodeQuery = true; else UrlDecodeQuery = false;
   RefererFilterPtr = CgiLibVar ("WWW_FORM_REFERER");
   if (!RefererFilterPtr[0])
      RefererFilterPtr = NULL;
   else
      for (cptr = RefererFilterPtr; *cptr; cptr++) *cptr = tolower(*cptr);
   cptr = CgiLibVar ("WWW_FORM_DECODEREFERER");
   if (cptr[0]) UrlDecodeReferer = true; else UrlDecodeReferer = false;
   UserAgentFilterPtr = CgiLibVar ("WWW_FORM_USERAGENT");
   if (!UserAgentFilterPtr[0])
      UserAgentFilterPtr = NULL;
   else
      for (cptr = UserAgentFilterPtr; *cptr; cptr++) *cptr = tolower(*cptr);
   ResponseFilterPtr = CgiLibVar ("WWW_FORM_RESPONSE");
   if (!ResponseFilterPtr[0])
      ResponseFilterPtr = CgiLibVar ("WWW_FORM_RESPONSELIST");
   if (!ResponseFilterPtr[0])
      ResponseFilterPtr = NULL;
   else
      for (cptr = ResponseFilterPtr; *cptr; cptr++) *cptr = tolower(*cptr);

   cptr = CgiLibVar ("WWW_FORM_VIEW");
   ShowLogFile = ShowProgressOnPage = false;
   if (strsame (cptr, "ALL", -1))
      ViewRecords = VIEW_ALL;
   else
   if (strsame (cptr, "ALL-LOG", -1))
   {
      ViewRecords = VIEW_ALL;
      ShowLogFile = true;
   }
   else
   if (strsame (cptr, "LOG", -1))
   {
      ViewRecords = VIEW_LOG;
      ShowLogFile = true;
   }
   else
   if (strsame (cptr, "MATCH", -1))
      ViewRecords = VIEW_MATCH;
   else
   if (strsame (cptr, "MATCH-LOG", -1))
   {
      ViewRecords = VIEW_MATCH;
      ShowLogFile = true;
   }
   else
   if (strsame (cptr, "NOMATCH", -1))
      ViewRecords = VIEW_NOMATCH;
   else
   if (strsame (cptr, "NOMATCH-LOG", -1))
   {
      ViewRecords = VIEW_NOMATCH;
      ShowLogFile = true;
   }
   else
   if (strsame (cptr, "PROGRESS", -1))
      ShowProgressOnPage = true;

   cptr = CgiLibVar ("WWW_FORM_ALL");
   if (cptr[0]) FilterOnAll = true;

   cptr = CgiLibVar ("WWW_FORM_LOOKUP");
   if (cptr[0]) LookupName = true;

   cptr = CgiLibVar ("WWW_FORM_MINSIZE");
   if (!cptr[0]) cptr = CgiLibVar ("WWW_FORM_MINSIZELIST");
   ResponseMinSize = atoi(cptr);
   if (isdigit(*cptr))
   {
      while (isdigit(*cptr)) cptr++;
      if (tolower(*cptr) == 'k') ResponseMinSize *= 1000;
      if (toupper(*cptr) == 'M') ResponseMinSize *= 1000000 ;
   }
   if (!ResponseMinSize)
      strcpy (ResponseMinSizeString, "<I>&lt;none&gt;</I>");
   else
   if (ResponseMinSize < 1000)
      sprintf (ResponseMinSizeString, "%d", ResponseMinSize); 
   else
   if (ResponseMinSize < 1000000)
      sprintf (ResponseMinSizeString, "%dk", ResponseMinSize / 1000); 
   else
      sprintf (ResponseMinSizeString, "%dM", ResponseMinSize / 1000000); 

   cptr = CgiLibVar ("WWW_FORM_MAXSIZE");
   if (!cptr[0]) cptr = CgiLibVar ("WWW_FORM_MAXSIZELIST");
   ResponseMaxSize = atoi(cptr);
   if (isdigit(*cptr))
   {
      while (isdigit(*cptr)) cptr++;
      if (tolower(*cptr) == 'k') ResponseMaxSize *= 1000;
      if (toupper(*cptr) == 'M') ResponseMaxSize *= 1000000;
   }
   if (!ResponseMaxSize)
      strcpy (ResponseMaxSizeString, "<I>&lt;none&gt;</I>");
   else
   if (ResponseMaxSize < 1000)
      sprintf (ResponseMaxSizeString, "%d", ResponseMaxSize); 
   else
   if (ResponseMaxSize < 1000000)
      sprintf (ResponseMaxSizeString, "%dk", ResponseMaxSize / 1000); 
   else
      sprintf (ResponseMaxSizeString, "%dM", ResponseMaxSize / 1000000); 

   /* WASD stream mode will stop each fflush() having carriage control added */
   CgiLibEnvironmentBinaryOut ();
   CgiLibResponseHeader (200, "text/html", "Script-Control: X-stream-mode\n");

   fprintf (stdout,
"<HTML>\n\
<HEAD>\n\
<META NAME=\"generator\" CONTENT=\"%s\">\n\
<META NAME=\"environment\" CONTENT=\"%s\">\n\
<META NAME=\"dna\" CONTENT=\"%s\">\n\
<TITLE>qdLogStats@%s</TITLE>\n\
</HEAD>\n\
<BODY>\n\
<FONT SIZE=+1><B><U>qdLogStats@%s</U></B></FONT><BR>\n\
<FONT SIZE=-3>&nbsp;</FONT>%s<BR>\n\
<FONT SIZE=-3>&nbsp;</FONT>%s\n",
      SoftwareID, CgiLibEnvironmentPtr, SearchDna,
      CgiServerNamePtr, CgiServerNamePtr,
      SOFTWAREID, StartDateTime);

   if (ViewRecords)
      fputs ("<P><HR SIZE=1 WIDTH=85% ALIGN=left NOSHADE>\n<PRE>", stdout);
   else
   {
      /* the CGI interface always provides hidden progress indication */
      ShowProgress = PROGRESS_RECORDS;
      if (ShowProgressOnPage)
         fputs ("<PRE>", stdout);
      else
         fputs ("<!-- just an indicator of progress ...\n", stdout);
   }
   fflush (stdout);

   ProgressWrap = PROGRESS_WRAP_CGI;

   /* just get the file portion of any file spec supplied */
   for (cptr = CgiFormLogFileSpecPtr; *cptr; cptr++);
   while (cptr > CgiFormLogFileSpecPtr && *cptr != ']') cptr--;
   if (*cptr == ']') cptr++;
   if (*cptr) CgiFormLogFileSpecPtr = cptr; 

   status = sys$setprv (1, &PrvMask, 0, 0);
   if (Debug) fprintf (stdout, "sys$setprv() %%X%08.08X\n", status);

   status = SearchFileSpec (CgiFormLogFileSpecPtr);

   sys$setprv (0, &PrvMask, 0, 0);

   if (ViewRecords)
      fputs ("</PRE>\n<HR SIZE=1 WIDTH=85% ALIGN=left NOSHADE>\n", stdout);
   else
   if (ShowProgressOnPage)
      fputs ("</PRE>\n", stdout);
   else
      fputs (" -->\n", stdout);

   if (VMSnok (status) && status != RMS$_NMF && status != RMS$_FNF)
      sprintf (VmsMessage,
"<TR><TH ALIGN=right><FONT COLOR=\"#ff0000\">ERROR</FONT>:&nbsp;</TH>\
<TD>%s<!-- %%X%08.08X --></TD></TR>\n\
<TR><TH HEIGHT=8></TH></TR>\n",
               SysGetMsg(status), status);
   else
      VmsMessage[0] = '\0';

   if (AccessProblemCount)
      sprintf (AccessProblemString, ", %d access failed", AccessProblemCount);
   else
      AccessProblemString[0] = '\0';

   if (LastModifiedIgnoredCount)
      sprintf (IgnoredString, ", %d ignored", LastModifiedIgnoredCount);
   else
      IgnoredString[0] = '\0';

   time (&TimeSeconds);
   TimeSeconds -= StartSeconds;
   DurationHr = TimeSeconds / 3600;
   TimeSeconds -= DurationHr * 3600;
   DurationMin = TimeSeconds / 60;
   DurationSec = TimeSeconds - DurationMin * 60;

   if (ByteCountTotal > 1000000000.0)
   {
      ByteCountTotal /= 1000000000.0;
      ByteStringPtr = "GBytes";
      sprintf (ByteString, "%.3f", ByteCountTotal);
   }
   else
   if (ByteCountTotal > 1000000.0)
   {
      ByteCountTotal /= 1000000.0;
      ByteStringPtr = "MBytes";
      sprintf (ByteString, "%.3f", ByteCountTotal);
   }
   else
   if (ByteCountTotal > 1000.0)
   {
      ByteCountTotal /= 1000.0;
      ByteStringPtr = "KBytes";
      sprintf (ByteString, "%.3f", ByteCountTotal);
   }
   else
   {
      ByteStringPtr = "Bytes";
      sprintf (ByteString, "%.0f", ByteCountTotal);
   }

   if (FilterOnAll)
   {
      if (!ClientFilterPtr)
         strcpy (ClientString, "<I>&lt;none&gt;</I>");
      else
         sprintf (ClientString, "&quot;%s&quot;&nbsp; (%d nomatch, %d match)",
                  (char*)CgiLibHtmlEscape (ClientFilterPtr, -1, NULL, 0),
                  ClientFilterRejectCount, ClientFilterAcceptCount);
      if (!AuthUserFilterPtr)
         strcpy (AuthUserString, "<I>&lt;none&gt;</I>");
      else
         sprintf (AuthUserString, "&quot;%s&quot;&nbsp; (%d nomatch, %d match)",
                  (char*)CgiLibHtmlEscape (AuthUserFilterPtr, -1, NULL, 0),
                  AuthUserFilterRejectCount, AuthUserFilterAcceptCount);
      if (!DateTimeFilterPtr)
         strcpy (DateTimeString, "<I>&lt;none&gt;</I>");
      else
         sprintf (DateTimeString, "&quot;%s&quot;&nbsp; (%d nomatch, %d match)",
                  (char*)CgiLibHtmlEscape (DateTimeFilterPtr, -1, NULL, 0),
                  DateTimeFilterRejectCount, DateTimeFilterAcceptCount);
      if (!MethodFilterPtr)
         strcpy (MethodString, "<I>&lt;none&gt;</I>");
      else
         sprintf (MethodString, "&quot;%s&quot;&nbsp; (%d nomatch, %d match)",
                  (char*)CgiLibHtmlEscape (MethodFilterPtr, -1, NULL, 0),
                  MethodFilterRejectCount, MethodFilterAcceptCount);
      if (!PathFilterPtr)
         strcpy (PathString, "<I>&lt;none&gt;</I>");
      else
         sprintf (PathString, "&quot;%s&quot;&nbsp; (%d nomatch, %d match)",
                  (char*)CgiLibHtmlEscape (PathFilterPtr, -1, NULL, 0),
                  PathFilterRejectCount, PathFilterAcceptCount);
      if (!QueryFilterPtr)
         strcpy (QueryString, "<I>&lt;none&gt;</I>");
      else
         sprintf (QueryString, "&quot;%s&quot;&nbsp; (%d nomatch, %d match)",
                  (char*)CgiLibHtmlEscape (QueryFilterPtr, -1, NULL, 0),
                  QueryFilterRejectCount, QueryFilterAcceptCount);
      if (!RefererFilterPtr)
         strcpy (RefererString, "<I>&lt;none&gt;</I>");
      else
         sprintf (RefererString, "&quot;%s&quot;&nbsp; (%d nomatch, %d match)",
                  (char*)CgiLibHtmlEscape (RefererFilterPtr, -1, NULL, 0),
                  RefererFilterRejectCount, RefererFilterAcceptCount);
      if (!RemoteIdentFilterPtr)
         strcpy (RemoteIdentString, "<I>&lt;none&gt;</I>");
      else
         sprintf (RemoteIdentString, "&quot;%s&quot;&nbsp; (%d nomatch, %d match)",
                  (char*)CgiLibHtmlEscape (RemoteIdentFilterPtr, -1, NULL, 0),
                  RemoteIdentFilterRejectCount, RemoteIdentFilterAcceptCount);
      if (!UserAgentFilterPtr)
         strcpy (UserAgentString, "<I>&lt;none&gt;</I>");
      else
         sprintf (UserAgentString, "&quot;%s&quot;&nbsp; (%d nomatch, %d match)",
                  (char*)CgiLibHtmlEscape (UserAgentFilterPtr, -1, NULL, 0),
                  UserAgentFilterRejectCount, UserAgentFilterAcceptCount);
      if (!ResponseFilterPtr)
         strcpy (ResponseString, "<I>&lt;all&gt;</I>");
      else
         sprintf (ResponseString, "&quot;%s&quot;&nbsp; (%d nomatch, %d match)",
                  (char*)CgiLibHtmlEscape (ResponseFilterPtr, -1, NULL, 0),
                  StatusFilterRejectCount, StatusFilterAcceptCount);
      if (ResponseMinSize)
      {
         for (cptr = ResponseMinSizeString; *cptr; cptr++);
         sprintf (cptr, "&nbsp; (%d nomatch, %d match)",
                  ResponseMinSizeRejectCount, ResponseMinSizeAcceptCount);
      }
      if (ResponseMaxSize)
      {
         for (cptr = ResponseMaxSizeString; *cptr; cptr++);
         sprintf (cptr, "&nbsp; (%d nomatch, %d match)",
                  ResponseMaxSizeRejectCount, ResponseMaxSizeAcceptCount);
      }
   }
   else
   {
      if (!ClientFilterPtr)
         strcpy (ClientString, "<I>&lt;none&gt;</I>");
      else
         sprintf (ClientString, "&quot;%s&quot;&nbsp; (%d nomatch)",
                  (char*)CgiLibHtmlEscape (ClientFilterPtr, -1, NULL, 0),
                  ClientFilterRejectCount);
      if (!AuthUserFilterPtr)
         strcpy (AuthUserString, "<I>&lt;none&gt;</I>");
      else
         sprintf (AuthUserString, "&quot;%s&quot;&nbsp; (%d nomatch)",
                  (char*)CgiLibHtmlEscape (AuthUserFilterPtr, -1, NULL, 0),
                  AuthUserFilterRejectCount);
      if (!DateTimeFilterPtr)
         strcpy (DateTimeString, "<I>&lt;none&gt;</I>");
      else
         sprintf (DateTimeString, "&quot;%s&quot;&nbsp; (%d nomatch)",
                  (char*)CgiLibHtmlEscape (DateTimeFilterPtr, -1, NULL, 0),
                  DateTimeFilterRejectCount);
      if (!MethodFilterPtr)
         strcpy (MethodString, "<I>&lt;none&gt;</I>");
      else
         sprintf (MethodString, "&quot;%s&quot;&nbsp; (%d nomatch)",
                  (char*)CgiLibHtmlEscape (MethodFilterPtr, -1, NULL, 0),
                  MethodFilterRejectCount);
      if (!PathFilterPtr)
         strcpy (PathString, "<I>&lt;none&gt;</I>");
      else
         sprintf (PathString, "&quot;%s&quot;&nbsp; (%d nomatch)",
                  (char*)CgiLibHtmlEscape (PathFilterPtr, -1, NULL, 0),
                  PathFilterRejectCount);
      if (!QueryFilterPtr)
         strcpy (QueryString, "<I>&lt;none&gt;</I>");
      else
         sprintf (QueryString, "&quot;%s&quot;&nbsp; (%d nomatch)",
                  (char*)CgiLibHtmlEscape (QueryFilterPtr, -1, NULL, 0),
                  QueryFilterRejectCount);
      if (!RefererFilterPtr)
         strcpy (RefererString, "<I>&lt;none&gt;</I>");
      else
         sprintf (RefererString, "&quot;%s&quot;&nbsp; (%d nomatch)",
                  (char*)CgiLibHtmlEscape (RefererFilterPtr, -1, NULL, 0),
                  RefererFilterRejectCount);
      if (!RemoteIdentFilterPtr)
         strcpy (RemoteIdentString, "<I>&lt;none&gt;</I>");
      else
         sprintf (RemoteIdentString, "&quot;%s&quot;&nbsp; (%d nomatch)",
                  (char*)CgiLibHtmlEscape (RemoteIdentFilterPtr, -1, NULL, 0),
                  RemoteIdentFilterRejectCount);
      if (!UserAgentFilterPtr)
         strcpy (UserAgentString, "<I>&lt;none&gt;</I>");
      else
         sprintf (UserAgentString, "&quot;%s&quot;&nbsp; (%d nomatch)",
                  (char*)CgiLibHtmlEscape (UserAgentFilterPtr, -1, NULL, 0),
                  UserAgentFilterRejectCount);

      if (!ResponseFilterPtr)
         strcpy (ResponseString, "<I>&lt;none&gt;</I>");
      else
         sprintf (ResponseString, "&quot;%s&quot;&nbsp; (%d nomatch)",
                  (char*)CgiLibHtmlEscape (ResponseFilterPtr, -1, NULL, 0),
                  StatusFilterRejectCount);
      if (ResponseMinSize)
      {
         for (cptr = ResponseMinSizeString; *cptr; cptr++);
         sprintf (cptr, "&nbsp; (%d nomatch)",
                  ResponseMinSizeRejectCount);
      }
      if (ResponseMaxSize)
      {
         for (cptr = ResponseMaxSizeString; *cptr; cptr++);
         sprintf (cptr, "&nbsp; (%d nomatch)",
                  ResponseMaxSizeRejectCount);
      }
   }

   fprintf (stdout,
"<P><TABLE CELLPADDING=0 CELLSPACING=3 BORDER=0>\n\
%s\
<TR><TH ALIGN=right>Log:&nbsp;</TH><TD>%s%s(%d matched%s%s)</TD></TR>\n\
%s\
<TR><TH HEIGHT=8></TH></TR>\n\
<TR><TH ALIGN=right>Client:&nbsp;</TH><TD>%s</TD></TR>\n\
<TR><TH ALIGN=right>Remote ID:&nbsp;</TH><TD>%s</TD></TR>\n\
<TR><TH ALIGN=right>Auth User:&nbsp;</TH><TD>%s</TD></TR>\n\
<TR><TH ALIGN=right>Date/Time:&nbsp;</TH><TD>%s</TD></TR>\n\
<TR><TH ALIGN=right>Method:&nbsp;</TH><TD>%s</TD></TR>\n\
<TR><TH ALIGN=right>Path:&nbsp;</TH><TD>%s</TD></TR>\n\
<TR><TH ALIGN=right>Query:&nbsp;</TH><TD>%s</TD></TR>\n\
<TR><TH ALIGN=right>Referer:&nbsp;</TH><TD>%s</TD></TR>\n\
<TR><TH ALIGN=right>User Agent:&nbsp;</TH><TD>%s</TD></TR>\n\
<TR><TH HEIGHT=8></TH></TR>\n\
<TR><TH ALIGN=right>Response:&nbsp;</TH><TD>%s</TD></TR>\n\
<TR><TH ALIGN=right>Size Min:&nbsp;</TH><TD>%s</TD></TR>\n\
<TR><TH ALIGN=right>Max:&nbsp;</TH><TD>%s</TD></TR>\n\
<TR><TH HEIGHT=8></TH></TR>\n\
<TR><TH ALIGN=right>Duration:&nbsp;</TH><TD>%02d:%02d:%02d</TD></TR>\n\
<TR><TH ALIGN=right>Records:&nbsp;</TH>\
<TD><NOBR>%d&nbsp; (%d suspect, %d common, %d combined)</NOBR></TD></TR>\n\
<TR><TH ALIGN=right>Requests:&nbsp;</TH><TD>%d</TD></TR>\n\
<TR><TH ALIGN=right>Methods:&nbsp;</TH>\
<TD><NOBR>CONNECT:%d&nbsp; DELETE:%d&nbsp; GET:%d&nbsp; HEAD:%d&nbsp; \
POST:%d&nbsp; PUT:%d&nbsp; ?:%d&nbsp;</NOBR></TD></TR>\n\
<TR><TH ALIGN=right>Responses:&nbsp;</TH>\
<TD><NOBR>1nn:%d&nbsp; 2nn:%d&nbsp; 3nn:%d&nbsp; 4nn:%d&nbsp; 5nn:%d&nbsp; \
?:%d&nbsp;</NOBR></TD></TR>\n\
<TR><TH ALIGN=right>%s:&nbsp;</TH><TD>%s</TD></TR>\n\
</TABLE>\n\
</BODY>\n\
</HTML>\n",
      VmsMessage,
      CgiFormLogFileSpecPtr,
      CgiFormLogFileSpecPtr[0] ? "&nbsp;&nbsp;" : "", FileCountTotal,
      IgnoredString,
      AccessProblemString,
      LastModifiedString,
      ClientString,
      RemoteIdentString,
      AuthUserString,
      DateTimeString,
      MethodString,
      PathString,
      QueryString,
      RefererString,
      UserAgentString,
      ResponseString,
      ResponseMinSizeString,
      ResponseMaxSizeString,
      DurationHr, DurationMin, DurationSec,
      RecordCountTotal, SuspectRecordCountTotal,
      CommonLogRecordCount, CombinedLogRecordCount,
      RequestCountTotal,
      MethodConnectCount,
      MethodDeleteCount,
      MethodGetCount,
      MethodHeadCount,
      MethodPostCount,
      MethodPutCount,
      MethodUnknownCount,
      StatusCodeCount[1],
      StatusCodeCount[2],
      StatusCodeCount[3],
      StatusCodeCount[4],
      StatusCodeCount[5],
      StatusCodeCount[0],
      ByteStringPtr, ByteString);
}

/****************************************************************************/
/*
*/ 

void CgiForm ()

{
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CgiForm()\n");

   /* make a guess at the log file defaults */
   if (CgiLibEnvironmentIsWasd())
      cptr = DEFAULT_LOGS_WASD;
   else
   if (CgiLibEnvironmentIsOsu())
      cptr = DEFAULT_LOGS_OSU;
   else
   if (CgiLibEnvironmentIsApache())
      cptr = DEFAULT_LOGS_APACHE;
   else
      cptr = DEFAULT_LOGS_FILESPEC;

   CgiLibEnvironmentPtr = CgiLibEnvironmentName();

   CgiScriptNamePtr = CgiLibVar ("WWW_SCRIPT_NAME");

   sys$fao (&StartDateTimeFaoDsc, NULL, &StartDateTimeDsc, 0);
   if (StartDateTime[0] == ' ') strcpy (StartDateTime, StartDateTime+1);

   CgiLibResponseHeader (200, "text/html");

   fprintf (stdout,
"<HTML>\n\
<HEAD>\n\
<META NAME=\"generator\" CONTENT=\"%s\">\n\
<META NAME=\"environment\" CONTENT=\"%s\">\n\
<META NAME=\"dna\" CONTENT=\"%s\">\n\
<TITLE>qdLogStats@%s</TITLE>\n\
</HEAD>\n\
<BODY>\n\
<FONT SIZE=+1><B><U>qdLogStats@%s</U></B></FONT><BR>\n\
<FONT SIZE=-3>&nbsp;</FONT>%s<BR>\n\
<FONT SIZE=-3>&nbsp;</FONT>%s\n\
<FORM METHOD=GET ACTION=\"%s\">\n\
<P><TABLE CELLPADDING=0 CELLSPACING=3 BORDER=0>\n\
<TR><TH ALIGN=right>Log&nbsp;Specification:&nbsp;</TH>\n\
<TD><INPUT TYPE=text NAME=logfilespec SIZE=30 VALUE=\"%s\">\
</TD></TR>\n\
<TR><TH ALIGN=right>Modified Since:&nbsp;</TH>\n\
<TD><INPUT TYPE=text NAME=logmodsince SIZE=11 VALUE=\"dd-mmm-%4.4s\">\
</TD></TR>\n\
<TR><TH ALIGN=right>Before:&nbsp;</TH>\n\
<TD><INPUT TYPE=text NAME=logmodbefore SIZE=11 VALUE=\"dd-mmm-%4.4s\">\
</TD></TR>\n\
<TR><TD HEIGHT=7></TD></TR>\n\
<TR><TH ALIGN=right>Client:&nbsp;</TH>\n\
<TD><INPUT TYPE=text NAME=client SIZE=30>\n\
&nbsp;<INPUT TYPE=checkbox NAME=lookup VALUE=1>\
&nbsp;<FONT SIZE=-1><I>lookup&nbsp;name</I></FONT>\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Remote ID:&nbsp;</TH>\n\
<TD><INPUT TYPE=text NAME=remoteid SIZE=30></TD></TR>\n\
<TR><TH ALIGN=right>Auth User:&nbsp;</TH>\n\
<TD><INPUT TYPE=text NAME=authuser SIZE=30></TD></TR>\n\
<TR><TH ALIGN=right>Date/Time:&nbsp;</TH>\n\
<TD><INPUT TYPE=text NAME=datetime SIZE=30 \
VALUE=\"dd/mmm/yyyy:hh:mm:ss &#177;hhmm\"></TD></TR>\n\
<TR><TH ALIGN=right>Method:&nbsp;</TH><TD>\n\
<SELECT NAME=methodlist>\n\
<OPTION VALUE=\"\" SELECTED>all\n\
<OPTION VALUE=CONNECT>CONNECT\n\
<OPTION VALUE=DELETE>DELETE\n\
<OPTION VALUE=GET>GET\n\
<OPTION VALUE=HEAD>HEAD\n\
<OPTION VALUE=POST>POST\n\
<OPTION VALUE=PUT>PUT\n\
</SELECT>\
&nbsp;<FONT SIZE=-1><I>or</I></FONT>&nbsp;\
<INPUT TYPE=text NAME=method SIZE=10>\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Path:&nbsp;</TH>\n\
<TD><INPUT TYPE=text NAME=path SIZE=30>\n\
&nbsp;<INPUT TYPE=checkbox NAME=decodePath VALUE=1>\
&nbsp;<FONT SIZE=-1><I>decode&nbsp;first</I></FONT>\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Query:&nbsp;</TH>\n\
<TD><INPUT TYPE=text NAME=query SIZE=30>\n\
&nbsp;<INPUT TYPE=checkbox NAME=decodeQuery VALUE=1>\
&nbsp;<FONT SIZE=-1><I>decode&nbsp;first</I></FONT>\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Referer:&nbsp;</TH>\n\
<TD><INPUT TYPE=text NAME=referer SIZE=30>\n\
&nbsp;<INPUT TYPE=checkbox NAME=decodeReferer VALUE=1>\
&nbsp;<FONT SIZE=-1><I>decode&nbsp;first</I></FONT>\n\
</TD></TR>\n\
<TR><TH ALIGN=right>User Agent:&nbsp;</TH>\n\
<TD><INPUT TYPE=text NAME=useragent SIZE=30></TD></TR>\n\
<TR><TD HEIGHT=8></TD></TR>\n\
<TR><TH ALIGN=right>Response:&nbsp;</TH><TD>\n\
<SELECT NAME=responselist>\n\
<OPTION VALUE=\"\" SELECTED>all\n\
<OPTION VALUE=2**>2nn - general success\n\
<OPTION VALUE=200>200 - success\n\
<OPTION VALUE=201>201 - created\n\
<OPTION VALUE=202>202 - accepted\n\
<OPTION VALUE=3**>3nn - client action\n\
<OPTION VALUE=300>300 - multiple choices\n\
<OPTION VALUE=301>301 - moved perm\n\
<OPTION VALUE=302>302 - moved temp\n\
<OPTION VALUE=304>304 - not modified\n\
<OPTION VALUE=4**>4nn - client error\n\
<OPTION VALUE=400>400 - bad request\n\
<OPTION VALUE=401>401 - authorization\n\
<OPTION VALUE=403>403 - forbidden\n\
<OPTION VALUE=404>404 - not found\n\
<OPTION VALUE=407>407 - proxy auth\n\
<OPTION VALUE=409>409 - conflict\n\
<OPTION VALUE=5**>5nn - server error\n\
<OPTION VALUE=500>500 - internal\n\
<OPTION VALUE=501>501 - not implemented\n\
<OPTION VALUE=502>502 - bad gateway\n\
<OPTION VALUE=503>503 - unavailable\n\
<OPTION VALUE=000>error (no status)\n\
</SELECT>\
&nbsp;<FONT SIZE=-1><I>or</I></FONT>&nbsp;\
<INPUT TYPE=text NAME=response SIZE=3>\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Size Min:&nbsp;</TH><TD>\n\
<SELECT NAME=minsizelist>\n\
<OPTION VALUE=\"\" SELECTED>none\n\
<OPTION VALUE=100>100\n\
<OPTION VALUE=250>250\n\
<OPTION VALUE=500>500\n\
<OPTION VALUE=1k>1k\n\
<OPTION VALUE=2.5k>2.5k\n\
<OPTION VALUE=5k>5k\n\
<OPTION VALUE=10k>10k\n\
<OPTION VALUE=25k>25k\n\
<OPTION VALUE=50k>50k\n\
<OPTION VALUE=100k>100k\n\
<OPTION VALUE=250k>250k\n\
<OPTION VALUE=500k>500k\n\
<OPTION VALUE=1M>1M\n\
<OPTION VALUE=10M>10M\n\
<OPTION VALUE=100M>100M\n\
</SELECT>\
&nbsp;<FONT SIZE=-1><I>or</I></FONT>&nbsp;\
<INPUT TYPE=text NAME=minsize SIZE=6>\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Max:&nbsp;</TH><TD>\n\
<SELECT NAME=maxsizelist>\n\
<OPTION VALUE=\"\" SELECTED>none\n\
<OPTION VALUE=100>100\n\
<OPTION VALUE=250>250\n\
<OPTION VALUE=500>500\n\
<OPTION VALUE=1k>1k\n\
<OPTION VALUE=2.5k>2.5k\n\
<OPTION VALUE=5k>5k\n\
<OPTION VALUE=10k>10k\n\
<OPTION VALUE=25k>25k\n\
<OPTION VALUE=50k>50k\n\
<OPTION VALUE=100k>100k\n\
<OPTION VALUE=250k>250k\n\
<OPTION VALUE=500k>500k\n\
<OPTION VALUE=1M>1M\n\
<OPTION VALUE=10M>10M\n\
<OPTION VALUE=100M>100M\n\
</SELECT>\
&nbsp;<FONT SIZE=-1><I>or</I></FONT>&nbsp;\
<INPUT TYPE=text NAME=maxsize SIZE=6>\n\
</TD></TR>\n\
<TR><TD HEIGHT=8></TD></TR>\n\
<TR><TH ALIGN=right>View:&nbsp;</TH><TD>\n\
<SELECT NAME=view>\n\
<OPTION VALUE=none SELECTED>none\n\
<OPTION VALUE=progress>progress\n\
<OPTION VALUE=match>matching records\n\
<OPTION VALUE=match-log>matching + log names\n\
<OPTION VALUE=nomatch>non-matching records\n\
<OPTION VALUE=nomatch-log>non-matching + log names\n\
<OPTION VALUE=all>all records\n\
<OPTION VALUE=all-log>all + log names\n\
<OPTION VALUE=log>log names\n\
</SELECT>\n\
&nbsp;<INPUT TYPE=checkbox NAME=all VALUE=1>\
<FONT SIZE=-1>&nbsp;<I>match all</I></FONT>\n\
</TD></TR>\n\
<TR><TH HEIGHT=4></TH></TR>\n\
<TR><TH></TH><TD>\n\
<INPUT TYPE=submit VALUE=\"Submit\">&nbsp;&nbsp;\
<INPUT TYPE=reset VALUE=\"reset\">\n\
</TD></TR>\n\
</TABLE>\n\
</FORM>\n\
</BODY>\n\
</HTML>\n",
      SoftwareID, CgiLibEnvironmentPtr, SearchDna,
      CgiServerNamePtr, CgiServerNamePtr,
      SOFTWAREID, StartDateTime,
      CgiScriptNamePtr, cptr,
      StartDateTime+7, StartDateTime+7);
}

/****************************************************************************/
/*
Generate and return a C/Unix time in seconds from a 'dd-mmm-yyyy' string (e.g.
10-AUG-2002).  Returns zero if string invalid.
*/ 

int DayMonthYear (char *String)

{
   static char  *MonthNames [] = { "JAN", "FEB", "MAR", "APR",
                                   "MAY", "JUN", "JUL", "AUG", 
                                   "SEP", "OCT", "NOV", "DEC", NULL };
   int  cnt, day, month, year, ctime;
   char  *cptr;
   char  mstring [256];
   struct tm  tmStruct;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DayMonthYear() |%s|\n", String);

   if (!String) return (0);

   for (cptr = String; *cptr; cptr++) *cptr = toupper(*cptr);
   day = year = 0;
   mstring[0] = '\0';
   cnt = sscanf (String, "%d-%3s-%d", &day, &mstring, &year);
   if (Debug) fprintf (stdout, "%d|%s|%d\n", day, mstring, year);
   if (cnt != 3) return (0);
   if (day < 1 || day > 31) return (0);
   if (year < 1990 || year > 2099) return (0);  /* now that's optimistic! */
   for (month = 0; MonthNames[month]; month++)
      if (strsame (mstring, MonthNames[month], -1)) break;

   memset (&tmStruct, 0, sizeof(tmStruct));
   tmStruct.tm_mday = day;
   tmStruct.tm_mon = month;
   tmStruct.tm_year = year - 1900;
   ctime = mktime (&tmStruct);
   if (Debug) fprintf (stdout, "ctime %u\n", ctime);

   if (ctime != -1) return (ctime);
   return (0);
}

/****************************************************************************/
/*
Turn the supplied string, for example "dd/aug/2002:dd:hh:mm +hhmm", into a
wilcarded date/time filter, such as "@@/aug/2002" and return a pointer to that
(sometimes modified) string.  If the string has not been changed from the
default value of "dd/mmm/yyyy:hh:mm:ss" then return NULL.
*/ 

char* DateTimeFilter (char *String)

{
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DateTimeFilter() |%s|\n", String);

   if (!String) return (NULL);
   if (strsame (String, "dd/mmm/yyyy:hh:mm:ss", 20)) return (NULL);

   cptr = String;
   if (*cptr == 'd') *cptr = '*';
   if (*cptr) cptr++;
   if (*cptr == 'd') *cptr = '*';
   if (*cptr) cptr++;
   if (*cptr != '/') return (String);
   if (*cptr) cptr++;

   if (*cptr == 'm') *cptr = '*';
   if (*cptr) cptr++;
   if (*cptr == 'm') *cptr = '*';
   if (*cptr) cptr++;
   if (*cptr == 'm') *cptr = '*';
   if (*cptr) cptr++;
   if (*cptr != '/') return (String);
   if (*cptr) cptr++;

   if (*cptr == 'y') *cptr = '*';
   if (*cptr) cptr++;
   if (*cptr == 'y') *cptr = '*';
   if (*cptr) cptr++;
   if (*cptr == 'y') *cptr = '*';
   if (*cptr) cptr++;
   if (*cptr == 'y') *cptr = '*';
   if (*cptr) cptr++;
   if (*cptr != ':') return (String);
   if (*cptr) cptr++;

   if (*cptr == 'h') *cptr = '*';
   if (*cptr) cptr++;
   if (*cptr == 'h') *cptr = '*';
   if (*cptr) cptr++;
   if (*cptr != ':') return (String);
   if (*cptr) cptr++;

   if (*cptr == 'm') *cptr = '*';
   if (*cptr) cptr++;
   if (*cptr == 'm') *cptr = '*';
   if (*cptr) cptr++;
   if (*cptr != ':') return (String);
   if (*cptr) cptr++;

   if (*cptr == 's') *cptr = '*';
   if (*cptr) cptr++;
   if (*cptr == 's') *cptr = '*';
   if (*cptr) cptr++;
   if (!isspace(*cptr)) return (String);
   if (*cptr) cptr++;

   if (*cptr != '+' && *cptr != '-' && *(unsigned char*)cptr != 177)
      return (String);
   *cptr++ = '*';
   if (*cptr == 'h') *cptr = '*';
   if (*cptr) cptr++;
   if (*cptr == 'h') *cptr = '*';
   if (*cptr) cptr++;
   if (*cptr == 'm') *cptr = '*';
   if (*cptr) cptr++;
   if (*cptr == 'm') *cptr = '*';
   if (*cptr) cptr++;
   if (*cptr) return (String);

   if (cptr > String && !*cptr) cptr--;
   while (cptr > String && !isalnum(*cptr)) cptr--;
   while (*cptr && *cptr != '*') cptr++;
   if (*cptr == '*') cptr++;
   *cptr = '\0';

   return (String);
}

/****************************************************************************/
/*
Search against the command line file specification, passing each file found to
be processed.
*/ 

int SearchFileSpec (char *FileSpec)

{
   int  status,
        FileNameLength,
        Length;
   char  FileName [256],
         ExpFileName [256];
   struct FAB  SearchFab;
   struct NAM  SearchNam;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "SearchFileSpec() |%s|%s|\n", SearchDna, FileSpec);

   /* initialize the file access block */
   SearchFab = cc$rms_fab;
   SearchFab.fab$l_dna = SearchDna;
   SearchFab.fab$b_dns = strlen(SearchDna);
   SearchFab.fab$l_fna = FileSpec;
   SearchFab.fab$b_fns = strlen(FileSpec);
   SearchFab.fab$l_nam = &SearchNam;

   SearchNam = cc$rms_nam;
   SearchNam.nam$l_esa = ExpFileName;
   SearchNam.nam$b_ess = sizeof(ExpFileName)-1;
   SearchNam.nam$l_rsa = FileName;
   SearchNam.nam$b_rss = sizeof(FileName)-1;

   if (VMSnok (status = sys$parse (&SearchFab, 0, 0))) return (status);

   while (VMSok (status = sys$search (&SearchFab, 0, 0)))
   {
      *SearchNam.nam$l_ver = '\0';
      if (Debug) fprintf (stdout, "FileName |%s|\n", FileName);

      ProcessLogFile (FileName);

      FileCountTotal++;
      *SearchNam.nam$l_ver = ';';
   }

   return (status);
}

/****************************************************************************/
/*
Open the log file, read and process each record from it, close the file!
*/ 

int ProcessLogFile (char *FileName)

{
   BOOL  accepted,
         ShowLog;
   int  status;
   char  *cptr;
   char  HtmlLine [MAX_LINE_LENGTH+(MAX_LINE_LENGTH/4)],
         Line [MAX_LINE_LENGTH],
         LineBuffer [MAX_LINE_LENGTH];
   FILE  *FilePtr;
   stat_t  statBuffer;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ProcessLogFile() |%s|\n", FileName);

   if (status = stat (FileName, &statBuffer) < 0)
   {
      if (Debug) fprintf (stdout, "stat() %d %%X%08.08X\n", status, vaxc$errno);
      AccessProblemCount++;
      status = vaxc$errno;
      return (status);
   }
   if (CtimeBefore && statBuffer.st_mtime > CtimeBefore)
   {
      LastModifiedIgnoredCount++;
      return (SS$_NORMAL);
   }
   if (CtimeSince && statBuffer.st_mtime < CtimeSince)
   {
      LastModifiedIgnoredCount++;
      return (SS$_NORMAL);
   }

   if (ShowProgress && !ViewRecords)
   {
      /* this works for both CLI and CGI (inside "<!-- -->"!) */
      if (ProgressCount == ProgressWrap)
      {
         ProgressCount = 0;
         fputc ('\n', stdout);
      }
      ProgressCount++;
      fputc ('+', stdout);
      fflush (stdout);
   }

   if (ShowLogFile)
      ShowLog = true;
   else
      ShowLog = false;

   RecordCount = 0;

   FilePtr = fopen (FileName, "r", "shr=get", "shr=put");
   if (!FilePtr)
   {
      status = vaxc$errno;
      if (Debug) fprintf (stdout, "fopen() %%X%08.08X\n", status);
      AccessProblemCount++;
      return (status);
   }

   while (fgets (Line, sizeof(Line), FilePtr))
   {
      RecordCount++;
      RecordCountTotal++;

      if (ViewRecords == VIEW_NONE)
      {
         /* again this works for both CLI and CGI (inside "<!-- -->"!) */
         if (ShowProgress && !(RecordCount % ShowProgress))
         {
            if (ProgressCount == ProgressWrap)
            {
               ProgressCount = 0;
               fputc ('\n', stdout);
            }
            ProgressCount++;
            fputc ('.', stdout);
            fflush (stdout);
         }
      }
      else
      if (ViewRecords == VIEW_ALL)
      {
         if (DoCliStats)
            OutputLogRecord (Line);
         else
         {
            CgiLibHtmlEscape (Line, -1, HtmlLine, sizeof(HtmlLine));
            OutputLogRecord (HtmlLine);
         }
      }
      else
         strcpy (LineBuffer, Line);

      if (ViewRecords == VIEW_LOG)
      {
         if (ShowLog)
         {   
            for (cptr = FileName; *cptr; cptr++);
            while (cptr > FileName && *cptr != ']' && *cptr != ':') cptr--;
            if (*cptr == ']' || *cptr == ':') cptr++;
            fprintf (stdout, "%s\n", cptr);
            ShowLog = false;
         }
      }

      if (Debug) fprintf (stdout, "|%s|\n", Line);
      accepted = ProcessLogRecord (Line);

      if (ViewRecords == VIEW_MATCH && accepted)
      {
         if (ShowLog)
         {   
            for (cptr = FileName; *cptr; cptr++);
            while (cptr > FileName && *cptr != ']' && *cptr != ':') cptr--;
            if (*cptr == ']' || *cptr == ':') cptr++;
            fprintf (stdout, "%s\n", cptr);
            ShowLog = false;
         }
         if (DoCliStats)
            OutputLogRecord (LineBuffer);
         else
         {
            CgiLibHtmlEscape (LineBuffer, -1, HtmlLine, sizeof(HtmlLine));
            OutputLogRecord (HtmlLine);
            fflush (stdout);
         }
      }
      else
      if (ViewRecords == VIEW_NOMATCH && !accepted)
      {
         if (ShowLog)
         {   
            for (cptr = FileName; *cptr; cptr++);
            while (cptr > FileName && *cptr != ']' && *cptr != ':') cptr--;
            if (*cptr == ']' || *cptr == ':') cptr++;
            fprintf (stdout, "%s\n", cptr);
            ShowLog = false;
         }
         if (DoCliStats)
            OutputLogRecord (LineBuffer);
         else
         {
            CgiLibHtmlEscape (LineBuffer, -1, HtmlLine, sizeof(HtmlLine));
            OutputLogRecord (HtmlLine);
            fflush (stdout);
         }
      }
   }

   fclose (FilePtr);

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
Output a single record from the log.  If client host name lookup is enabled
then check whether the client field is a dotted-decimal host address.  If it is
then lookup the host name and output that instead of the address.
*/

void OutputLogRecord (char *StringPtr)

{
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "OutputLogRecord()\n");

   if (!LookupName)
   {
      fputs (StringPtr, stdout);
      return;
   }

   for (cptr = StringPtr; *cptr && (isdigit(*cptr) || *cptr == '.'); cptr++);
   if (!isspace(*cptr))
   {
      fputs (StringPtr, stdout);
      return;
   }

   if (LookupNamePtr)
   {
      fputs (LookupNamePtr, stdout);
      fputs (cptr, stdout);
      return;
   }

   cptr = LookupHostName (&StringPtr);
   if (!cptr)
   {
      fputs (StringPtr, stdout);
      return;
   }

   fputs (cptr, stdout);
   fputs (StringPtr, stdout);
}

/*****************************************************************************/
/*
Process a single record (line) from the log file.
Common: 'client rfc891 authuser date/time request status bytes'
Combined: 'client rfc891 authuser date/time request status bytes referer agent'
*/

BOOL ProcessLogRecord (char *Line)

{
   BOOL  RejectRecord;
   int  ByteCount;
   char  *cptr, *sptr,
         *AuthUserPtr,
         *ClientPtr,
         *BytesPtr,
         *DateTimePtr,
         *QueryStringPtr,
         *RefererPtr,
         *RemoteIdentPtr,
         *RequestPtr,
         *StatusPtr,
         *UserAgentPtr;
   char  Scratch [MAX_LINE_LENGTH];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ProcessLogRecord()\n");

   LookupNamePtr = NULL;

   cptr = Line;

   /* client */
   ClientPtr = cptr;
   while (*cptr && !isspace(*cptr)) cptr++;
   if (*cptr) *cptr++ = '\0';

   /* remote ident (RFC891) */
   while (*cptr && isspace(*cptr)) cptr++;
   RemoteIdentPtr = cptr;
   while (*cptr && !isspace(*cptr)) cptr++;
   if (*cptr) *cptr++ = '\0';

   /* authorized user */
   while (*cptr && isspace(*cptr)) cptr++;
   AuthUserPtr = cptr;
   while (*cptr && !isspace(*cptr)) cptr++;
   if (*cptr) *cptr++ = '\0';

   /* date/time */
   while (*cptr && *cptr != '[') cptr++;
   if (*cptr) cptr++;
   DateTimePtr = cptr;
   while (*cptr && *cptr != ']') cptr++;
   if (*cptr) *cptr++ = '\0';

   /* request ("method path?query") */
   while (*cptr && *cptr != '\"') cptr++;
   if (*cptr) cptr++;
   RequestPtr = cptr;
   while (*cptr && *cptr != '\"') cptr++;
   if (*cptr) *cptr++ = '\0';

   /* HTTP response status */
   while (*cptr && isspace(*cptr)) cptr++;
   StatusPtr = cptr;
   while (*cptr && !isspace(*cptr)) cptr++;
   if (*cptr) *cptr++ = '\0';

   /* bytes transmitted */
   while (*cptr && isspace(*cptr)) cptr++;
   BytesPtr = cptr;
   while (*cptr && !isspace(*cptr)) cptr++;
   if (*cptr) *cptr++ = '\0';

   /* referer (only for COMBINED) */
   RefererPtr = NULL;
   while (*cptr && isspace(*cptr)) cptr++;
   if (*cptr == '\"')
   {
      cptr++;
      RefererPtr = cptr;
      while (*cptr && *cptr != '\"') cptr++;
   }
   else
   if (*cptr)
   {
      RefererPtr = cptr;
      while (*cptr && !isspace(*cptr)) cptr++;
   }
   if (*cptr) *cptr++ = '\0';

   /* user agent (only for COMBINED) */
   UserAgentPtr = NULL;
   while (*cptr && isspace(*cptr)) cptr++;
   if (*cptr == '\"')
   {
      cptr++;
      UserAgentPtr = cptr;
      while (*cptr && *cptr != '\"') cptr++;
   }
   else
   if (*cptr)
   {
      UserAgentPtr = cptr;
      while (*cptr && !isspace(*cptr)) cptr++;
   }
   if (*cptr) *cptr++ = '\0';

   if (!ClientPtr[0] ||
       !DateTimePtr[0] ||
       !RequestPtr[0] ||
       !StatusPtr[0] ||
       !BytesPtr)
   {
      SuspectRecordCountTotal++;
      return (false);
   }

   if (!RefererPtr && !UserAgentPtr)
      CommonLogRecordCount++;
   else
      CombinedLogRecordCount++;

   RejectRecord = false;

   if (ClientFilterPtr)
   {
      if (LookupName)
      {
         cptr = LookupHostName (&ClientPtr);
         if (cptr) ClientPtr = cptr;
      }
      if (FilterThisOut (ClientPtr, ClientFilterPtr))
      {
         ClientFilterRejectCount++;
         if (!FilterOnAll) return (false);
         RejectRecord = true;
      }
      else
         ClientFilterAcceptCount++;
   }
   if (RemoteIdentFilterPtr)
   {
      if (FilterThisOut (RemoteIdentPtr, RemoteIdentFilterPtr))
      {
         RemoteIdentFilterRejectCount++;
         if (!FilterOnAll) return (false);
         RejectRecord = true;
      }
      else
         RemoteIdentFilterAcceptCount++;
   }
   if (AuthUserFilterPtr)
   {
      if (FilterThisOut (AuthUserPtr, AuthUserFilterPtr))
      {
         AuthUserFilterRejectCount++;
         if (!FilterOnAll) return (false);
         RejectRecord = true;
      }
      else
         AuthUserFilterAcceptCount++;
   }
   if (DateTimeFilterPtr)
   {
      if (FilterThisOut (DateTimePtr, DateTimeFilterPtr))
      {
         DateTimeFilterRejectCount++;
         if (!FilterOnAll) return (false);
         RejectRecord = true;
      }
      else
         DateTimeFilterAcceptCount++;
   }
   if (MethodFilterPtr)
   {
      for (cptr = RequestPtr; *cptr && *cptr != ' '; cptr++);
      /* terminate path at query string (should be necessary) */
      if (!*cptr) cptr = NULL;
      if (cptr) *cptr = '\0';
      if (FilterThisOut (RequestPtr, MethodFilterPtr))
      {
         MethodFilterRejectCount++;
         if (!FilterOnAll) return (false);
         RejectRecord = true;
      }
      else
         MethodFilterAcceptCount++;
      if (cptr) *cptr = ' ';
   }
   if (PathFilterPtr)
   {
      /* skip over the method */
      for (cptr = RequestPtr; *cptr && *cptr != ' '; cptr++);
      while (*cptr && *cptr == ' ') cptr++;
      /* find the start of any query string */
      for (sptr = cptr; *sptr && *sptr != '?'; sptr++);
      /* terminate path at query string if necessary */
      if (*sptr)
         QueryStringPtr = sptr + 1;
      else
         QueryStringPtr = sptr = NULL;
      if (sptr) *sptr = '\0';
      if (UrlDecodePath && *cptr)
      {
         strcpy (Scratch, cptr);
         cptr = Scratch;
         CgiLibUrlDecode (cptr);
      }
      if (FilterThisOut (cptr, PathFilterPtr))
      {
         PathFilterRejectCount++;
         if (!FilterOnAll) return (false);
         RejectRecord = true;
      }
      else
         PathFilterAcceptCount++;
      /* return request to pristine condition :^) */
      if (sptr) *sptr = '?';
   }
   if (QueryFilterPtr)
   {
      if (PathFilterPtr)
      {
         /* path has already been parsed, query string has been determined */
         if (QueryStringPtr)
            cptr = QueryStringPtr;
         else
            cptr = "";
      }
      else
      {
         /* find the start of any query string */
         for (cptr = RequestPtr; *cptr && *cptr != '?'; cptr++);
         if (*cptr) cptr++;
      }
      if (UrlDecodeQuery && *cptr)
      {
         strcpy (Scratch, cptr);
         cptr = Scratch;
         CgiLibUrlDecode (cptr);
      }
      if (FilterThisOut (cptr, QueryFilterPtr))
      {
         QueryFilterRejectCount++;
         if (!FilterOnAll) return (false);
         RejectRecord = true;
      }
      else
         QueryFilterAcceptCount++;
   }
   if (RefererFilterPtr)
   {
      /* if the record did not have a COMBINED referer field then reject */
      if ((cptr = RefererPtr) && UrlDecodeReferer)
      {
         strcpy (Scratch, cptr);
         cptr = Scratch;
         CgiLibUrlDecode (cptr);
      }
      if (!cptr || FilterThisOut (cptr, RefererFilterPtr))
      {
         RefererFilterRejectCount++;
         if (!FilterOnAll) return (false);
         RejectRecord = true;
      }
      else
         RefererFilterAcceptCount++;
   }
   if (ResponseFilterPtr)
   {
      if (!cptr || FilterThisOut (StatusPtr, ResponseFilterPtr))
      {
         StatusFilterRejectCount++;
         if (!FilterOnAll) return (false);
         RejectRecord = true;
      }
      else
         StatusFilterAcceptCount++;
   }
   if (UserAgentFilterPtr)
   {
      /* if the record did not have a COMBINED user agent field then reject */
      if (!UserAgentPtr || FilterThisOut (UserAgentPtr, UserAgentFilterPtr))
      {
         UserAgentFilterRejectCount++;
         if (!FilterOnAll) return (false);
         RejectRecord = true;
      }
      else
         UserAgentFilterAcceptCount++;
   }

   ByteCount = atoi(BytesPtr);
   if (ResponseMinSize)
   {
      if (ByteCount < ResponseMinSize || !isdigit(*BytesPtr))
      {
         ResponseMinSizeRejectCount++;
         if (!FilterOnAll) return (false);
         RejectRecord = true;
      }
      else
         ResponseMinSizeAcceptCount++;
   }
   if (ResponseMaxSize)
   {
      if (ByteCount > ResponseMaxSize || !isdigit(*BytesPtr))
      {
         ResponseMaxSizeRejectCount++;
         if (!FilterOnAll) return (false);
         RejectRecord = true;
      }
      else
         ResponseMaxSizeAcceptCount++;
   }

   if (RejectRecord) return (false);

   RequestCountTotal++;
   ByteCountTotal += (float)ByteCount;
   switch (StatusPtr[0])
   {
      case '1' : StatusCodeCount[1]++; break;
      case '2' : StatusCodeCount[2]++; break;
      case '3' : StatusCodeCount[3]++; break;
      case '4' : StatusCodeCount[4]++; break;
      case '5' : StatusCodeCount[5]++; break;
      default  : StatusCodeCount[0]++;
   }
   switch (*(unsigned short*)RequestPtr)
   {
      /* unfiltered it still will be uppercase, filtered lower-case */
      case 'CO' : case 'co' : MethodConnectCount++; break;
      case 'DE' : case 'de' : MethodDeleteCount++; break;
      case 'GE' : case 'ge' : MethodGetCount++; break;
      case 'HE' : case 'he' : MethodHeadCount++; break;
      case 'PO' : case 'po' : MethodPostCount++; break;
      case 'PU' : case 'pu' : MethodPutCount++; break;
      default   : MethodUnknownCount++;
   }

   return (true);
}

/*****************************************************************************/
/*
Using decc$match_wild() match the pattern against the string.  Parses out the
"+{}" and "-{}" kludges from composite expressions and does the appropriate
boolean processing to return a result.  Expects the pattern already to be all
lower-case, as this is a case-less matching.
*/

BOOL FilterThisOut
(
char *StringPtr,
char *PatternPtr
)
{
   BOOL  Matched,
         NegateMatch;
   char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "FilterThisOut()\n");

   NegateMatch = false;
   for (cptr = StringPtr; *cptr; cptr++) *cptr = tolower(*cptr);
   cptr = PatternPtr;
   while (cptr && *cptr)
   {
      if (*(unsigned short*)cptr == '+{')
      {
         NegateMatch = false;
         cptr += 2;
      }
      else
      if (*(unsigned short*)cptr == '-{')
      {
         NegateMatch = true;
         cptr += 2;
      }
      else
         return (!decc$match_wild (StringPtr, PatternPtr));

      sptr = cptr;
      while (*cptr && *cptr != '}') cptr++;
      if (!*cptr) cptr = NULL;
      if (cptr) *cptr = '\0';
      Matched = decc$match_wild (StringPtr, sptr);
      if (cptr) *cptr++ = '}';
      if (NegateMatch) Matched = !Matched;
      if (!Matched) return (true);
   }

   return (false);
}

/*****************************************************************************/
/*
Lookup the host name by address.  The names are kept in a local, fixed length
cache (array).
*/

char* LookupHostName (char **StringPtrPtr)

{
   static int  NameCacheCount,
               CurrentIndex;
   static struct
   {
      int  HitCount,
           IpAddress,
           LookupFailCount;
      char  *NamePtr;
   } NameCache [MAX_LOOKUP_CACHE];

   int  idx,
        IpAddress;
   char  ch;
   char  *cptr;
   struct hostent  *HostEntryPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "LookupHostName() |%s|\n", *StringPtrPtr);

   if (!LookupName) return (NULL);

   for (cptr = *StringPtrPtr;
        *cptr && (isdigit(*cptr) || *cptr == '.');
        cptr++);
   if (*cptr && !isspace(*cptr)) return (NULL);

   ch = *cptr;
   *cptr = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", *StringPtrPtr);
   IpAddress = inet_addr (*StringPtrPtr);
   *cptr = ch;
   if (Debug) fprintf (stdout, "%08.08X\n", IpAddress);
   if (IpAddress == -1) return (NULL);

   if (IpAddress == NameCache[CurrentIndex].IpAddress &&
       !NameCache[CurrentIndex].LookupFailCount)
   {
      NameCache[CurrentIndex].HitCount++;
      *StringPtrPtr = cptr;
      return (NameCache[CurrentIndex].NamePtr);
   }

   for (CurrentIndex = 0; CurrentIndex < NameCacheCount; CurrentIndex++)
   {
      if (IpAddress != NameCache[CurrentIndex].IpAddress) continue;
      if (!NameCache[CurrentIndex].NamePtr)
      {
         /* this is an address hit but with no host name resolved */
         if (NameCache[CurrentIndex].LookupFailCount >= MAX_LOOKUP_FAILS)
            return (NULL);
         /* attempts to resolve name are still within limit, have another go */
         break;
      }
      /* this is an address hit with a resolved name, just return the name */
      NameCache[CurrentIndex].HitCount++;
      *StringPtrPtr = cptr;
      return (NameCache[CurrentIndex].NamePtr);
   }

   HostEntryPtr = gethostbyaddr (&IpAddress, sizeof(IpAddress), AF_INET);

   if (CurrentIndex >= NameCacheCount)
   {
      /* the IP address was not found in the cache */
      if (NameCacheCount < MAX_LOOKUP_CACHE-1)
      {
         /* there is still room in the cache */
         CurrentIndex = NameCacheCount++;
      }
      else
      {
         /* find the first entry with the lowest hit rate */
         CurrentIndex = 0;
         for (idx = 0; idx < NameCacheCount; idx++)
         {
            if (NameCache[CurrentIndex].IpAddress && !NameCache[idx].HitCount)
               continue;
            if (NameCache[idx].HitCount >= NameCache[CurrentIndex].HitCount)
               continue; 
            CurrentIndex = idx;
         }
         NameCache[CurrentIndex].HitCount =
            NameCache[CurrentIndex].LookupFailCount = 0;
         free (NameCache[CurrentIndex].NamePtr);
      }
      NameCache[CurrentIndex].IpAddress = IpAddress;
   }

   if (!HostEntryPtr)
   {
      /* the name didn't resolve */
      NameCache[CurrentIndex].LookupFailCount++;
      return (NameCache[CurrentIndex].NamePtr = NULL);
   }

   NameCache[CurrentIndex].LookupFailCount = 0;
   NameCache[CurrentIndex].NamePtr = malloc(strlen(HostEntryPtr->h_name)+1);
   strcpy (NameCache[CurrentIndex].NamePtr, HostEntryPtr->h_name);

   *StringPtrPtr = cptr;
   return (NameCache[CurrentIndex].NamePtr);
}

/*****************************************************************************/
/*
Get "command-line" parameters, whether from the command-line or from a
configuration symbol or logical containing the equivalent.
*/

GetParameters ()

{
   static char  CommandLine [256];
   static unsigned long  Flags = 0;

   int  status,
        SkipParameters;
   unsigned short  Length;
   char  ch;
   char  *aptr, *cptr, *clptr, *sptr;
   $DESCRIPTOR (CommandLineDsc, CommandLine);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "GetParameters()\n");

   if (!(clptr = getenv ("QDLOGSTATS$PARAM")))
   {
      /* get the entire command line following the verb */
      if (VMSnok (status =
          lib$get_foreign (&CommandLineDsc, 0, &Length, &Flags)))
         exit (status);
      (clptr = CommandLine)[Length] = '\0';
   }

   /* if CSWS (Apache) */
   if (getenv ("APACHE$SHARED_SOCKET"))
   {
      /* look for something non-space outside of quotes */
      for (cptr = clptr; *cptr; cptr++)
      {
         if (isspace(*cptr)) continue;
         if (*cptr != '\"') break;
         cptr++;
         while (*cptr && *cptr != '\"') cptr++;
         if (*cptr) cptr++;
      }
      /* if nothing outside of quotes then ignore the command line */
      if (!*cptr) return;
   }

   /* if OSU environment then skip P1, P2, P3 */
   if (getenv ("WWWEXEC_RUNDOWN_STRING"))
      SkipParameters = 3;
   else
      SkipParameters = 0;

   aptr = NULL;
   ch = *clptr;
   for (;;)
   {
      if (aptr && *aptr == '/') *aptr = '\0';
      if (!ch) break;

      *clptr = ch;
      if (Debug) fprintf (stdout, "clptr |%s|\n", clptr);
      while (*clptr && isspace(*clptr)) *clptr++ = '\0';
      aptr = clptr;
      if (*clptr == '/') clptr++;
      while (*clptr && !isspace (*clptr) && *clptr != '/')
      {
         if (*clptr != '\"')
         {
            clptr++;
            continue;
         }
         cptr = clptr;
         clptr++;
         while (*clptr)
         {
            if (*clptr == '\"')
               if (*(clptr+1) == '\"')
                  clptr++;
               else
                  break;
            *cptr++ = *clptr++;
         }
         *cptr = '\0';
         if (*clptr) clptr++;
      }
      ch = *clptr;
      if (*clptr) *clptr = '\0';
      if (Debug) fprintf (stdout, "aptr |%s|\n", aptr);
      if (!*aptr) continue;

      if (SkipParameters)
      {
         SkipParameters--;
         continue;
      }

      if (strsame (aptr, "/ALL", -1))
      {
         FilterOnAll = true;
         continue;
      }
      if (strsame (aptr, "/AUTHUSER=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         for (AuthUserFilterPtr = cptr; *cptr; cptr++) *cptr = tolower(*cptr);
         continue;
      }
      if (strsame (aptr, "/BEFORE=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         LastModifiedBeforePtr = cptr;
         continue;
      }
      if (strsame (aptr, "/CLIENT=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         for (ClientFilterPtr = cptr; *cptr; cptr++) *cptr = tolower(*cptr);
         continue;
      }
      if (strsame (aptr, "/DBUG", -1))
      {
         Debug = true;
         continue;
      }
      if (strsame (aptr, "/DATETIME=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         for (DateTimeFilterPtr = cptr; *cptr; cptr++) *cptr = tolower(*cptr);
         continue;
      }
      if (strsame (aptr, "/DECODE=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         if (!*cptr)
         {
            UrlDecodePath = UrlDecodeQuery = UrlDecodeReferer = true;
            continue;
         }
         if (*cptr == '(') cptr++;
         while (*cptr && *cptr != ')')
         {
            while (*cptr == ',') cptr++;
            if (*cptr) cptr++;
            if (toupper(*cptr) == 'P')
               UrlDecodePath = true;
            else
            if (toupper(*cptr) == 'Q')
               UrlDecodeQuery = true;
            else
            if (toupper(*cptr) == 'R')
               UrlDecodeReferer = true;
            else
            {
               for (sptr = cptr; *sptr && *sptr != ',' && *sptr != ')'; sptr++);
               *sptr = '\0';
               fprintf (stdout, "%%%s-E-INVKEYW, invalid keyword\n \\%s\\\n",
                        Utility, cptr);
               exit (STS$K_ERROR | STS$M_INHIB_MSG);
            }
            while (*cptr && *cptr != ',' && *cptr != ')') cptr++;
         }
         continue;
      }
      if (strsame (aptr, "/LOG", -1))
      {
         ShowLogFile = true;
         continue;
      }
      if (strsame (aptr, "/LOOKUP", -1))
      {
         LookupName = true;
         continue;
      }
      if (strsame (aptr, "/METHOD=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         for (MethodFilterPtr = cptr; *cptr; cptr++) *cptr = tolower(*cptr);
         continue;
      }
      if (strsame (aptr, "/OUTPUT=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         OutputPtr = cptr;
         continue;
      }
      if (strsame (aptr, "/PATH=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         for (PathFilterPtr = cptr; *cptr; cptr++) *cptr = tolower(*cptr);
         continue;
      }
      if (strsame (aptr, "/PROGRESS=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         ShowProgress = atoi(cptr);
         if (ShowProgress <= 0) ShowProgress = PROGRESS_RECORDS;
         continue;
      }
      if (strsame (aptr, "/QUERY=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         for (QueryFilterPtr = cptr; *cptr; cptr++) *cptr = tolower(*cptr);
         continue;
      }
      if (strsame (aptr, "/REFERER=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         for (RefererFilterPtr = cptr; *cptr; cptr++) *cptr = tolower(*cptr);
         continue;
      }
      if (strsame (aptr, "/REMOTEID=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         for (RemoteIdentFilterPtr = cptr; *cptr; cptr++) *cptr = tolower(*cptr);
         continue;
      }
      if (strsame (aptr, "/RESPONSE", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         for (ResponseFilterPtr = cptr; *cptr; cptr++) *cptr = tolower(*cptr);
         continue;
      }
      if (strsame (aptr, "/SINCE=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         LastModifiedSincePtr = cptr;
         continue;
      }
      if (strsame (aptr, "/SIZE=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         while (*cptr == '=') cptr++;
         while (*cptr == '(') cptr++;
         while (*cptr)
         {
            if (strsame (cptr, "MIN=", 4))
            {
               cptr += 4;
               ResponseMinSize = atoi(cptr);
               while (isdigit(*cptr)) cptr++;
               if (tolower(*cptr) == 'k') ResponseMinSize *= 1000;
               if (toupper(*cptr) == 'M') ResponseMinSize *= 1000000;
            }
            else
            if (strsame (cptr, "MAX=", 4))
            {
               cptr += 4;
               ResponseMaxSize = atoi(cptr);
               while (isdigit(*cptr)) cptr++;
               if (tolower(*cptr) == 'k') ResponseMaxSize *= 1000;
               if (toupper(*cptr) == 'M') ResponseMaxSize *= 1000000;
            }
            else
            {
               fprintf (stdout, "%%%s-E-IVKEYW, unrecognized keyword\n \\%s\\\n",
                        Utility, cptr);
               exit (STS$K_ERROR | STS$M_INHIB_MSG);
            }
            while (*cptr && *cptr != ',' && *cptr != ')') cptr++;
            while (*cptr == ',' || *cptr == ')') cptr++;
         }
         continue;
      }
      if (strsame (aptr, "/SOFTWAREID", 4) ||
          strsame (aptr, "/VERSION", 4))
      {
         fprintf (stdout, "%%%s-I-SOFTWAREID, %s\n%s\n",
                  Utility, SoftwareID, CopyrightInfo);
         exit (SS$_NORMAL);
      }
      if (strsame (aptr, "/USERAGENT=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         for (UserAgentFilterPtr = cptr; *cptr; cptr++) *cptr = tolower(*cptr);
         continue;
      }
      if (strsame (aptr, "/VIEW=", 4))
      {
         ViewRecords = VIEW_MATCH;
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (!*cptr) continue;
         cptr++;
         if (strsame (cptr, "ALL", 3))
            ViewRecords = VIEW_ALL;
         else
         if (strsame (cptr, "LOG", 3))
            ViewRecords = VIEW_LOG;
         else
         if (strsame (cptr, "MATCH", 3))
            ViewRecords = VIEW_MATCH;
         else
         if (strsame (cptr, "NOMATCH", 3))
            ViewRecords = VIEW_NOMATCH;
         else
         {
            fprintf (stdout, "%%%s-E-IVKEYW, unrecognized keyword\n \\%s\\\n",
                     Utility, cptr);
            exit (STS$K_ERROR | STS$M_INHIB_MSG);
         }
         continue;
      }

      if (*aptr == '/')
      {
         fprintf (stdout, "%%%s-E-IVQUAL, unrecognized qualifier\n \\%s\\\n",
                  Utility, aptr+1);
         exit (STS$K_ERROR | STS$M_INHIB_MSG);
      }

      if (!LogFileSpec[0])
      {
         sptr = LogFileSpec;
         for (cptr = aptr; *cptr; *sptr++ = toupper(*cptr++));
         *sptr = '\0';
         continue;
      }

      fprintf (stdout, "%%%s-E-MAXPARM, too many parameters\n \\%s\\\n",
               Utility, aptr);
      exit (STS$K_ERROR | STS$M_INHIB_MSG);
   }
}

/*****************************************************************************/
/*
Get the value of a system-level (only) logical name.  Returns a pointer to
dynamic storage containing the null-terminated string, or NULL if a problem.
A *little* like 'getenv()' but only from the system logical table - serves much
the same purpose anyway.
*/

char* SysTrnLnmLnmSystem (char *LogicalName)

{
   static unsigned short  Length;
   static char  LogicalValue [256];
   static $DESCRIPTOR (LogicalNameDsc, "");
   static $DESCRIPTOR (LnmSystemDsc, "LNM$SYSTEM");
   static struct {
      short int  buf_len;
      short int  item;
      void  *buf_addr;
      unsigned short  *ret_len;
   } LnmItems [] =
   {
      { sizeof(LogicalValue)-1, LNM$_STRING, LogicalValue, &Length },
      { 0,0,0,0 }
   };

   int  status;
   char  *sptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SysTrnLnmLnmSystem() |%s|\n", LogicalName);

   LogicalNameDsc.dsc$w_length = strlen(LogicalName);
   LogicalNameDsc.dsc$a_pointer = LogicalName;

   status = sys$trnlnm (0, &LnmSystemDsc, &LogicalNameDsc, 0, &LnmItems);
   if (Debug) fprintf (stdout, "sys$trnlnm() %%X%08.08X\n", status);
   if (VMSnok (status)) return (NULL);

   LogicalValue[Length] = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", LogicalValue);
   if (!(sptr = calloc (1, Length+1))) exit (vaxc$errno);
   memcpy (sptr, LogicalValue, Length+1);
   return (sptr);
}

/*****************************************************************************/
/*
Return a pointer to the message string corresponding to the supplied VMS status
value.
*/

char* SysGetMsg (int VmsStatus)

{
   static char  Message [256];

   int  status;
   short int  Length;
   $DESCRIPTOR (MessageDsc, Message);

   /*********/
   /* begin */
   /*********/

   Message[0] = '\0';
   if (VmsStatus)
   {
      /* text component only */
      sys$getmsg (VmsStatus, &Length, &MessageDsc, 1, 0);
      Message[Length] = '\0';
   }
   if (Message[0])
      return (Message);
   else
      return ("(internal error)");
}
 
/*****************************************************************************/
/*
Because it can be installed with privileges (for CLI usage) ...
*/ 

void NeedsPrivilegedAccount ()

{
   static unsigned long  PrivAcctMask [2] = { PRV$M_SETPRV | PRV$M_SYSPRV, 0 };

   static long  Pid = -1;
   static unsigned long  JpiAuthPriv [2];

   static struct
   {
      unsigned short  buf_len;
      unsigned short  item;
      void  *buf_addr;
      void  *ret_len;
   }
      JpiItems [] =
   {
      { sizeof(JpiAuthPriv), JPI$_AUTHPRIV, &JpiAuthPriv, 0 },
      {0,0,0,0}
   };

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NeedsPrivilegedAccount()\n");

   status = sys$getjpiw (0, &Pid, 0, &JpiItems, 0, 0, 0);
   if (VMSnok (status)) exit (status);

   if (!(JpiAuthPriv[0] & PrivAcctMask[0])) exit (SS$_NOSYSPRV);
}

/****************************************************************************/
/*
Does a case-insensitive, character-by-character string compare and returns 
true if two strings are the same, or false if not.  If a maximum number of 
characters are specified only those will be compared, if the entire strings 
should be compared then specify the number of characters as 0.
*/ 

BOOL strsame
(
char *sptr1,
char *sptr2,
int  count
)
{
   /*********/
   /* begin */
   /*********/

   while (*sptr1 && *sptr2)
   {
      if (toupper (*sptr1++) != toupper (*sptr2++)) return (false);
      if (count)
         if (!--count) return (true);
   }
   if (*sptr1 || *sptr2)
      return (false);
   else
      return (true);
}

/*****************************************************************************/
/*
*/

int ShowHelp ()

{
   fputs (
"Usage for Quick and Dirty LOG STATisticS\n\
\n\
$ QDLOGSTATS <log-file-spec> [<filters>] [<other qualifiers>]\n\
\n\
Utility to extract very elementary statistics from Web server common/combined\n\
format log files.  A number of filters allow subsets of the log contents to be\n\
selected using simple wildcard expressions.  Strings are NOT case-sensitive.\n\
Optionally, log file records can be viewed as processed, or a simple progress\n\
indicator can be displayed (\"+\" for each file, \".\" per 1000 records thereof).\n\
\n\
/ALL /AUTHUSER=filter /BEFORE=time /CLIENT=filter /DATETIME=filter\n\
/DECODE[=keyword] /LOG /LOOKUP /METHOD=filter /OUTPUT=file /PATH=filter\n\
/PROGRESS[=integer] /QUERY=filter /REFERER=filter /RESPONSE=filter\n\
/SINCE=time /SIZE=[MIN=,MAX=] /SOFTWAREID /USERAGENT=filter\n\
/VIEW[=MATCH(D)|ALL|NOMATCH|LOG]\n\
\n\
$ QDLOGSTATS == \"$dir:QDLOGSTATS\"\n\
$ QDLOGSTATS HT_LOGS:*.LOG /PROGRESS\n\
$ QDLOGSTATS HT_LOGS:*.LOG /VIEW /PATH=\"/wasd/*.zip\"\n\
$ QDLOGSTATS HT_LOGS:*NOV*ACCESS* /PATH=\"/CGI-BIN/*\" /QUERY=\"-{-}\"\n\
$ QDLOGSTATS HT_LOGS:*ACCESS*.LOG /METHOD=POST /DATE=\"*/NOV/2000/*\"\n\
$ QDLOGSTATS HT_LOGS:*.LOG /SINCE=01-FEB-2002 /USERAGENT=*MOZILLA*X11*\n\
\n", stdout);

   return (SS$_NORMAL);
}

/*****************************************************************************/

